#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include <ctype.h>

#define ASCIIZ '\0'
 
 /* 
  * Processor dependent defines.  
  * Define one of:
  * PCXT   : PC stuff but not vectra.
  * HP150
  * or
  * HPUNIX
  *
  * The following are derived from the above:
  * PC for all PC's and HP150
  * HPTERM is for HP terminal escape sequences
  */

/* #define HPUNIX */
#define PCXT
/* #define HP150 */

#ifdef HPUNIX
#define HPTERM
#endif

#ifdef HP150
#define HPTERM
#define PC
#endif

#ifdef PCXT
#define PC
#endif
 
#ifdef PC
#include <dos.h>
#endif

#ifdef HPTERM
#define SPECIAL '\033'
#endif

#ifndef HPTERM
#define SPECIAL ASCIIZ
#endif

#ifdef HPUNIX
#include <termio.h>
#include <fcntl.h>
#include <signal.h>
#define byte unsigned char
#endif

#define BELL '\007'
#define SPACE ' '
#define BACKSPACE '\010'
 
#ifdef HPTERM
#define F1KEY 'p'
#define F2KEY 'q'
#define F3KEY 'r'
#define F4KEY 's'
#define F5KEY 't'
#define F6KEY 'u'
#define F7KEY 'v'
#define F8KEY 'w'
#define DOWNKEY 'B'
#define UPKEY 'A'
#define LEFTKEY 'D'
#define ROLLKEY 'T'
#define DELKEY 'P'
#define NEXTKEY 'U'
#define PREVKEY 'V'
#define HOMEKEY 'h'
#endif

#ifndef HPTERM
#define F1KEY '\073'
#define F2KEY '\074'
#define F3KEY '\075'
#define F4KEY '\076'
#define F5KEY '\077'
#define F6KEY '\100'
#define F7KEY '\101'
#define F8KEY '\102'
#define F9KEY '\103'
#define F10KEY '\104'
#define HOMEKEY '\107'
#define UPKEY '\110'
#define LEFTKEY '\113'
#define DOWNKEY '\120'
#define ROLLKEY '\121'
#define DELKEY '\123'
#define NEXTKEY '\124'
#define PREVKEY '\125'
#endif

/* screen addresses ... */ 
#define LEFTSIDE 2
#define START_X 10
#define STARTL_Y 9
#define STARTT_Y 3
#define STARTZ_Y 4
#define STARTY_Y 5
#define STARTX_Y 7
#define S_REGS_X 44
#define S_REGS_Y 3
#define BASE_X 2
#define BASE_Y 1
#define FK_X 40
#define FK_Y 15
#define INV_X 11
#define INV_Y 1
#define TMODE_X 21
#define TMODE_Y 1
#define INTLINEX 2
#define INTLINEY 11
#define ERRORX 2
#define ERRORY 12

/* COMMANDS (in addition to 'normal' keys ... */
#define ASCII '\200'
#define FLOAT '\201'
#define BIN   '\202'
#define OCT   '\203'
#define DEC   '\204'
#define HEX   '\205'

#define INV   '\206'
#define SIN   '\207'
#define COS   '\210'
#define TAN   '\211'
#define SINH  '\212'
#define COSH  '\213'
#define TANH  '\214'
#define LOGE  '\215'
#define LOG10 '\216'

#define SQR   '\217'
#define ROOT  '\220'
#define CUBE  '\221'
#define CROOT '\222'
#define FRC   '\223'
#define INT   '\224'

#define SUM   '\225'
#define SUM0  '\226'
#define SUMR  '\227'
#define MEAN  '\230'
#define S_DEV '\231'
#define NSTAT '\232'

#define NFIN  '\233'
#define INTST '\234'
#define PMT   '\235'
#define PVAL  '\236'
#define FVAL  '\237'
#define ENG   '\240'

#define ROLLDOWN '\300'
#define CLX      '\301'
#define REFRESH '\302'
#define QUIT '\303'

#define F1 '\340'
#define F2 '\341'
#define F3 '\342'
#define F4 '\343'
#define F5 '\344'
#define F6 '\345'
#define F7 '\346'
#define F8 '\347'
#define F9 '\350'
#define F10 '\351'

/* ... END OF COMMANDS */

#define TRUE 1
#define FALSE 0
#define BOOLEAN int

#define NUMREGS 10
#define NUMFK 5

static  long x, y, z, t, l, reg[NUMREGS];
static  double xf, yf, zf, tf, lf, regf[NUMREGS], pi, d_to_r, r_to_d;
static  BOOLEAN dot_legal, e_legal, entering, lift_needed, invert, degree;
static  BOOLEAN last_was_fin_key, line_drawing;
static  char inbuf[80], *inptr, outbuf[80];
static  char  zero_string[] = "                               0";
static  char blank_string[] = "                                ";
static  int base, fk_set, msg_flag, max_digits;
static  jmp_buf save;

static struct {
		char label[9];
		char function;
	      } func[NUMFK][8] =
	     
{ " Ascii  ", ASCII,
  " Float  ", FLOAT,
  " Engin. ", ENG,
  " Binary ", BIN,
  " Octal  ", OCT,
  " Dec    ", DEC,
  " Hex    ", HEX,
  " Refresh", REFRESH,

  "Inverse ", INV,
  " Sin    ", SIN,
  " Cos    ", COS,
  " Tan    ", TAN,
  " Log E  ", LOGE,
  " Log 10 ", LOG10,
  "        ", ASCIIZ,
  " Refresh", REFRESH,

  " Square ", SQR,
  " Root   ", ROOT,
  " Cube   ", CUBE,
  " C root ", CROOT,
  " FRC    ", FRC,
  " INT    ", INT,
  "        ", ASCIIZ,
  " Refresh", REFRESH,

  " Sum    ", SUM,
  " Mean   ", MEAN,
  " S Dev  ", S_DEV,
  " N      ", NSTAT,
  " Clear  ", SUM0,
  " Sum-   ", SUMR,
  "        ", ASCIIZ,
  " Refresh", REFRESH,

  " N      ", NFIN,
  " I      ", INTST,
  " P Val  ", PVAL,
  " PMT    ", PMT,
  " F Val  ", FVAL,
  "        ", ASCIIZ,
  "        ", ASCIIZ,
  " Refresh", REFRESH,
};

#ifdef HPUNIX
/*
   set stdin (file desc 0) tty characteristics to a state enabling
   unbuffered io, ready for 'read_char'. The old tty status is returned
   as the function value.
*/

static struct termio old;
static state = 1; /* Used to ensure uncanon is not called twice, thus
		     overwriting the 'old' struct */

extern void uncanon()

{
  unsigned short mask = ISIG;
  struct termio new;
  int fd;

  if (state) {
    fd = fileno(stdin);
    if (ioctl(fd, TCGETA, &old) < 0)
      fprintf(stderr, "Uncanon: ioctl get failed\n");
    ioctl(fd, TCGETA, &new);
    new.c_lflag &= mask;
    new.c_cc[4] = 1; /* set MIN = 1 */
    new.c_cc[5] = 0; /* set TIME = 0 */
    if (ioctl(fd, TCSETAW, &new) < 0)
      fprintf(stderr, "Uncanon: ioctl set failed\n");
    state = 0;
  }
}

/* set stdin (file desc = 0) to a stty status defined in the structure
   passed in 'old' ... usually set by the function 'uncanonical'. */

extern void canon()

{
  int fd;

  if (state != 1)
    {
    fd = fileno(stdin);
    ioctl(fd, TCSETAW, &old);
    }
  state = 1;
}
#endif

print_string(s)
char *s;

{
  while (*s != ASCIIZ)
    {
    put_a_char(*s);
    s++;
    }
}

draw_line(s)
char *s;

{
#ifdef HPTERM
static BOOLEAN char_set_selected = FALSE;

  if (!char_set_selected)
    {
    print_string("\033)B");
    char_set_selected = TRUE;
    }
  print_string(s);
#endif

#ifndef HPTERM
  BOOLEAN shift_in;
  int len;
  char *p;
  
  shift_in = FALSE;
  p = s;
  while (*p != ASCIIZ)
    {
    if ((*p == '') || (*p == ''))
      shift_in = !shift_in;
    else if (!shift_in)
      put_a_char(*p);
    else
      switch (toupper(*p))
        {
	case 'Q':
	  put_a_char('\311');
	  break;
	case 'W':
	  put_a_char('\273');
	  break;
	case ';':
	  put_a_char('\315');
	  break;
	case '*':
	  put_a_char('\327');
	  break;
	case '3':
	  put_a_char('\313');
	  break;
	case ':':
	  put_a_char('\272');
	  break;
	case 'A':
	  put_a_char('\310');
	  break;
	case '4':
	  put_a_char('\312');
	  break;
	case 'S':
	  put_a_char('\274');
	  break;
	case '"':
	  put_a_char('\266');
	  break;
	case '!':
	  put_a_char('\307');
	  break;
	case 'R':
	  put_a_char('\332');
	  break;
	case ',':
	  put_a_char('\304');
	  break;
	case 'T':
	  put_a_char('\277');
	  break;
	case '.':
	  put_a_char('\263');
	  break;
	case 'F':
	  put_a_char('\300');
	  break;
	case 'G':
	  put_a_char('\331');
	  break;
	case '\n':
	  put_a_char('\n');
	  put_a_char('\r');
	  break;
	default:
	  put_a_char(*p);
	  break;
	}
    p++;  
    }  
#endif
}
 
print_attr(s, attr)
char *s;
int attr;

/* attr = 0 for normal, 
	  1 for inverse, 
	  2 for underlined (blue on black on CGA */

{
#ifndef HPTERM
  int x, y;
  union REGS inregs, outregs;

  inregs.h.ah = 3; /* get cursor position */
  inregs.h.bh = 0; /* page number */
  int86(0x10, &inregs, &outregs);
  x = outregs.h.dl;
  y = outregs.h.dh;
  
  inregs.h.ah = 9; /* put character & attribute */
  if (attr == 0)
    inregs.h.bl = 0x07; /* attribute = normal */
  else if (attr == 1)
    inregs.h.bl = 0x70; /* attribute = inverse video */
  else
    inregs.h.bl = 0x01; /* attribute = underlined */
  inregs.h.bh = 0;
  inregs.x.cx = 1;
  while (*s != ASCIIZ)
    {
    inregs.h.al = *(s++);
    int86(0x10, &inregs, &outregs);
    gotoxy(++x, y);
    }
#endif

#ifdef HPTERM
  if (attr == 0)
    print_string("\033&d@");
  else if (attr == 1)
    print_string("\033&dB");
  else
    print_string("\033&dD");
  print_string(s);
  print_string("\033&d@");
#endif
}

msg(s)
char *s;

{
  gotoxy(ERRORX, ERRORY);
  strcpy(outbuf, s);
  while (strlen(outbuf) < 40)
    strcat(outbuf, " ");
  print_attr(outbuf, 1);
  dispnums();
  msg_flag = strlen(s);
}

put_a_char(c)
char c;
 
{
#ifdef HP150
/* This code uses high level DOS function to write a char.
   Takes about 25 secs for 10 refreshes of hpcalc screen.
   Use BIOS call instead ... */

  union REGS inregs, outregs;
 
  inregs.h.ah = 2;
  inregs.h.dl = c;
  intdos(&inregs, &outregs);
  if (c == '\n')
    put_a_char('\r');
#endif

#ifdef PCXT
/* Use ROM BIOS call to write a char.
   Takes about 16 secs for 10 refreshes of hpcalc screen. */
  union REGS inregs, outregs;
  
  inregs.h.ah = 0x0E;
  inregs.h.al = c;
  inregs.h.bl = 0;
  int86(0x10, &inregs, &outregs);
  if (c == '\n')
    put_a_char('\r');
#endif

#ifdef HPUNIX
  putchar(c);
#endif
}

clear_display()
 
{
#ifdef HPTERM
  print_string("\033h\033J");
#endif

#ifdef PCXT
  union REGS inregs, outregs;
 
  inregs.h.ah = 15;
  int86(16, &inregs, &outregs);
  inregs.h.ah = 0;
  inregs.h.al = outregs.h.al;
  int86(16, &inregs, &outregs);
#endif
}
 
gotoxy(x, y)
int x, y;
 
{
#ifdef HPTERM
  sprintf(outbuf, "\033&a%dc%dR", x, y);
  print_string(outbuf);
#endif

#ifndef HPTERM
  union REGS inregs, outregs;
 
  inregs.h.ah = 2;
  inregs.h.dh = y;
  inregs.h.dl = x;
  inregs.h.bh = 0;
  int86(16, &inregs, &outregs);
#endif
}
 
char get_a_char(d)
char *d;
 
{
  /* get a character with no echo, no wait. If special, return SPECIAL
     with special value in d. Else d = 0.
  */
 
#ifdef PC
  union REGS inregs, outregs;
  char c;
 
  *d = ASCIIZ;
  inregs.h.ah = 7;
  gotoxy(START_X + strlen(inbuf), STARTX_Y);
  intdos(&inregs, &outregs);
  if (outregs.h.al == SPECIAL)
    {
    inregs.h.ah = 7; /* get char from k/b no echo */
    intdos(&inregs, &outregs);
    *d = outregs.h.al;
    return(SPECIAL);
    }
  c = outregs.h.al;
  if (base != 1)
    c = toupper(c);
  return(c);
#endif

#ifdef HPUNIX
  char c;
  int old;

  gotoxy(START_X + strlen(inbuf), STARTX_Y);
  fflush(stdout);
  c = getchar();
  if (base != 1)
    c = toupper(c);
  if (c == SPECIAL)  /* do a no wait read - if ESCAPE then this will return 0 */
    {
    old = fcntl(0, F_GETFL);
    fcntl(0, F_SETFL, old | O_NDELAY);
    *d = getchar();
    fcntl(0, F_SETFL, old);
    }
  return(c);
#endif
}
 
echo_char(c)
char c;
/* echo 'c' back to the screen and log in the input buffer */
 
{
  long temp;
  double tempf;
 
  if (c == BACKSPACE)
    {
    if (entering && (inptr > inbuf))
      {
      put_a_char(c);
      put_a_char(SPACE);
      put_a_char(c);
      inptr--;
      if ((*inptr == '.') && (base < 0))
        dot_legal = TRUE;
      if ((*inptr == 'e') && (base < 0))
        e_legal = TRUE;
      *inptr = ASCIIZ;
      }
    else
      put_a_char(BELL);
    }
  else
    {
    if (!entering)
      {
      if (lift_needed)
	{
        temp = x;
        tempf = xf;
        push(temp, tempf);
	dispnums();
	}
      lift_needed = TRUE;
      entering = TRUE;
      dot_legal = TRUE;
      e_legal = TRUE;
      inptr = inbuf;
      *inptr = ASCIIZ;
      gotoxy(START_X, STARTX_Y);
      print_string(blank_string);
      gotoxy(START_X, STARTX_Y);
      }
    if (inptr < inbuf + max_digits)
      {
      put_a_char(c);
      *(inptr++) = c;
      *inptr = ASCIIZ;
      }
    else
      put_a_char(BELL);
    }
}
 
fmt_bin(str, x)
char *str;
long x;
 
{
  char inverted[50], reverted[50], *s, *t;
  BOOLEAN negative;
 
  if (x < 0)
    {
    negative = TRUE;
    x &= 0x7FFFFFFF;
    }
  else
    negative = FALSE;
  s = inverted;
  while (x != 0)
    {
    *(s++) = '0' + (char) (x % 2);
    x /= 2;
    }
  *s = ASCIIZ;
  if (negative)
    {
    while (strlen(inverted) < 31)
      {
      *(s++) = '0';
      *s = ASCIIZ;
      }
    *(s++) = '1';
    }
  if (s == inverted) /* x==0 */
    strcpy(str, zero_string);
  else
    {
    t = reverted;
    while (s != inverted)
      *(t++) = *(--s);
    *t = ASCIIZ;
    sprintf(str, "%32.32s", reverted);
    }
}
 
fmt_base(s, x, xf)
char *s;
long x;
double xf;
 
{
  char format[20], d;
  union frig {
  	long j; 
	byte c[5];
	     } frigger;
  int i;
  
  switch (base)
    {
    case 1:
      frigger.j = x;
      frigger.c[4] = ASCIIZ;

      strcpy(s, blank_string);
      for (i = 0; i < 4; i++)
        {
	d = frigger.c[i];
#ifdef HPTERM
        if (d < 32)
#endif
#ifndef HPTERM
	if ((d == ASCIIZ) || 
	    (d == BACKSPACE) ||
	    (d == BELL) ||
	    (d == '\n') ||
	    (d == '\r'))
#endif
	  {
	  d = SPACE;
	  }
        s[28 + i] = d;
	}
      break;
      
    case 2:
      fmt_bin(s, x);
      break;
    case 8:
      if (x == 0L)
        strcpy(s, zero_string);
      else
        sprintf(s, "%32.lo", x);
      break;
    case 10:
      if (x == 0L)
        strcpy(s, zero_string);
      else
        {
        sprintf(s, "%32.ld", x);
	}
      break;
    case 16:
      if (x == 0L)
        strcpy(s, zero_string);
      else
        sprintf(s, "%32.lx", x);
      break;
    default:
      if ((base < -9) || (fabs(xf) > 1.0E30))
	{
	i = -base;
	if (i > 9)
	  i -= 10;
        sprintf(format, "%%32.%d", i);
        strcat(format, "lg");
        sprintf(s, format, xf);
	}
      else
	{
        sprintf(format, "%%32.%d", -base);
        strcat(format, "lf");
        sprintf(s, format, xf);
	}
      break;
    }
}

print_base(x, xf)
long x;
double xf;

{
  fmt_base(outbuf, x, xf);
  print_string(outbuf);
}
 
dispreg(i)
int i;

{
    gotoxy(S_REGS_X, S_REGS_Y + i);
    sprintf(outbuf, "%d ", i);
    fmt_base(&outbuf[2], reg[i], regf[i]);
    if (line_drawing)
      print_attr(outbuf, 2);
    else
      print_string(outbuf);
}

dispregs()

{
  int i;

  gotoxy(S_REGS_X + 1, 1);
  print_string("REGISTERS");
  for (i = 0; i < NUMREGS; i++)
    dispreg(i);
}

dispnums()
 
{
  gotoxy(START_X, STARTT_Y);
  print_base(t, tf);
  gotoxy(START_X, STARTZ_Y);
  print_base(z, zf);
  gotoxy(START_X, STARTY_Y);
  print_base(y, yf);
  gotoxy(START_X, STARTX_Y);
  if (entering)
    print_string(inbuf);
  else
    print_base(x, xf);
  gotoxy(START_X, STARTL_Y);
  print_base(l, lf);
}
 
prinbase()

{
  gotoxy(BASE_X, BASE_Y);
  switch (base)
    {
    case 1:
      print_string("Ascii   ");
      break;

    case 2:
      print_string("Binary  ");
      break;

    case 8:
      print_string("Octal   ");
      break;

    case 10:
      print_string("Decimal ");
      break;

    case 16:
      print_string("Hex     ");
      break;

    default:
      if (base < -9)
        {
        sprintf(outbuf, "Engin. %d", -(base + 10));
	print_string(outbuf);
	}
      else
        {
        sprintf(outbuf, "Float  %d", -base);
	print_string(outbuf);
	}
      break;
    }
}

print_fk(i)
int i;

{
  int j;
  
  invert = FALSE;  
  print_inv();

#ifndef HPTERM
  gotoxy(FK_X, FK_Y);
  print_string("Function Keys:\n");
  for (j = 0; j < 8; j += 2)
    {
    gotoxy(FK_X, FK_Y + 1 + j/2);
    sprintf(outbuf, "%d:%s", j + 1, func[i][j].label);
    print_string(outbuf);
    sprintf(outbuf, "  %d:%s", j + 2, func[i][j + 1].label);
    print_string(outbuf);
    }
  gotoxy(FK_X, FK_Y + 5);
  print_string("9: Prev    10: Next   ");
#endif
    
#ifdef HPTERM
  gotoxy(FK_X, FK_Y);
  print_string("Use NEXT/PREV to scroll function keys");
  gotoxy(FK_X, FK_Y + 2);
  print_string("If no line drawing set, use \'dcalc -l\'");
  for (j = 0; j < 8; j++)
    {
    sprintf(outbuf, "\033&f0a%dk8d0L%s", j + 1, func[i][j].label);
    print_string(outbuf);
    }
  print_string("\033&jB");
#endif
}

print_inv()

{
	gotoxy(INV_X, INV_Y);
        if (invert)
	  print_string("Arc() ");
	else
	  print_string("      ");
}

print_deg()

{
  gotoxy(TMODE_X, TMODE_Y);
  if (degree)
    print_string("Degrees");
  else
    print_string("Radians");
}

print_frame()

{
draw_line("q;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;3;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;w\n", 2);
draw_line(":                                          :                                  :\n");
draw_line("!,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,*,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,\"\n");
draw_line(":                                          :                                  :\n");
draw_line(":                                          :                                  :\n");
draw_line(":                                          :                                  :\n");
draw_line(":        r,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,t:                                  :\n");
draw_line(":        .                                .:                                  :\n");
draw_line(":        f,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,,g:                                  :\n");
draw_line(":                                          :                                  :\n");
draw_line(":                                          :                                  :\n");
draw_line(":                                          :                                  :\n");
draw_line(":                                          :                                  :\n");
draw_line("a;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;4;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;s\n");
}

display()
 
{
  int i;
  
  clear_display();
  if (line_drawing)
    print_frame();
  prinbase();
  print_deg();
  print_inv();
  gotoxy(LEFTSIDE, STARTT_Y);
  print_string("     T:");
  gotoxy(LEFTSIDE, STARTZ_Y);
  print_string("     Z:");
  gotoxy(LEFTSIDE, STARTY_Y);
  print_string("     Y:");
  gotoxy(LEFTSIDE, STARTX_Y);
  print_string("     X:");
  gotoxy(LEFTSIDE, STARTL_Y);
  print_string("Last X:");

  dispnums();
  dispregs();
  
  gotoxy(LEFTSIDE, 14);
  print_string("L: Last X     +-*/");
  gotoxy(LEFTSIDE, 15);
  print_string("S: Store      %: x% of y");
  gotoxy(LEFTSIDE, 16);
  print_string("R: Recall     Y: y**x");
  gotoxy(LEFTSIDE, 17);
  print_string("Home   : Quit #: y mod x");
  gotoxy(LEFTSIDE, 18);
  print_string("Del Chr: CLX  &: x AND y");
  gotoxy(LEFTSIDE, 19);
  print_string("M: Deg/Rad    |: x OR y");
  gotoxy(LEFTSIDE, 20);
  print_string(",: CHS        ^: x XOR y");
  gotoxy(LEFTSIDE, 21);
  print_string(" : ROLL DN    ~: compl x");
  gotoxy(LEFTSIDE, 22);
  print_string("X: x<>y     >,<: y l/r shift x bits");
  print_fk(fk_set);
}
 
long asc2int(s)
char *s;

{
  union frig {
  	long j; 
	char c[5];
	     } frigger;
  int i;

  for (i = 0; i < 4 - strlen(s); i++)
    frigger.c[i] = ASCIIZ;
  for (; i < 4; i++)
    frigger.c[i] = *(s++);
  return(frigger.j);
}

long bin2int(s)
char *s;
 
{
  char *t;
  long accum;
 
  accum = 0;
  t = s;
  while (*t != ASCIIZ)
    accum = 2 * accum + (*(t++) - '0');
  return(accum);
}
 
stop_entering()
 
{
  long temp;
  double tempf;
 
  if (entering)
    {
    entering = FALSE;
    switch (base)
      {
      case 1:
	temp = asc2int(inbuf);
	tempf = temp;
	break;
      case 2:
        temp = bin2int(inbuf);
        tempf = temp;
        break;
      case 8:
        sscanf(inbuf, "%lo", &temp);
        tempf = temp;
        break;
      case 10:
        sscanf(inbuf, "%ld", &temp);
        tempf = temp;
        break;
      case 16:
        sscanf(inbuf, "%lx", &temp);
        tempf = temp;
        break;
      default:
	tempf = atof(inbuf);
	if (fabs(tempf) <= 0x7fffffff)
	  temp = tempf;
	else
	  temp = 0;
        break;
      }
    x = temp;
    xf = tempf;
    }
}

push(inx, inxf)
long inx;
double inxf;
 
{
  stop_entering();
  t = z;
  tf = zf;
  z = y;
  zf = yf;
  y = x;
  yf = xf;
  x = inx;
  xf = inxf;
  lift_needed = TRUE;
}
 
pop(outx, outxf)
long *outx;
double *outxf;
 
{
  stop_entering();
  *outx = x;
  *outxf = xf;
  x = y;
  xf = yf;
  y = z;
  yf = zf;
  z = t;
  zf = tf;
}
 
int get_int(s)
char *s;
 
{
  char c, d, buf[80];
  int t, retval;
 
  gotoxy(INTLINEX, INTLINEY);
  print_string(s);
  c = get_a_char(&d);
  if (isdigit(c))
    retval = c - '0';
  else
    {
    put_a_char(BELL);
    retval = -1;
    }
  gotoxy(INTLINEX, INTLINEY);
  strcpy(buf, s);
  for (t = 0; t < strlen(s) ; t++)
    buf[t] = SPACE;
  print_string(buf);
  return(retval);
}
 
save_x()
 
{
  stop_entering();
  l = x;
  lf = xf;
}
 
char map_key_to_func(c, d)
char c, d;

{
  char ret_val;
  
  ret_val = c;
  if (c == SPECIAL)
    {
    if ((d >= F1KEY) && (d <= F8KEY))
      ret_val = func[fk_set][d - F1KEY].function;
    else switch (d)
      {
      case ROLLKEY:
      case DOWNKEY:
	ret_val = ROLLDOWN;
	break;
	
      case UPKEY:
	ret_val = '\n';
	break;

#ifndef HPTERM
      case F10KEY:
#endif
      case NEXTKEY:
	ret_val = NEXTKEY;
	break;

#ifndef HPTERM
      case F9KEY:
#endif
      case PREVKEY:
	ret_val = PREVKEY;
	break;

      case DELKEY:
        ret_val = CLX;
	break;

      case HOMEKEY:
	ret_val = QUIT;
	break;

      default:
        break;
      }
    }
  return(ret_val);
}

long round(f)
double f;

{
  BOOLEAN negative;  
  long ret;
  
  negative = FALSE;
  if (f < 0.0)
    {
    negative = TRUE;
    f = -f;
    }
  ret = floor(f + 0.5);
  if (negative)
    ret = -ret;
  return(ret);
}
 
process(c)
char *c;
 
{
  long tempx, tempy;
  double tempxf, tempyf, n, interest, pval, pmt, fval, guess, epsilon;
  int i, count;
  char *c_ptr, c_buf[10], d;
  byte dummy;
  BOOLEAN this_was_fin_key;

  this_was_fin_key = FALSE;
  dummy = *c;
  if ((base == 1)  && (dummy > '\037') && (dummy < '\200'))
    {
    echo_char(*c);
    return(0);
    }
    switch (*c)
      {
      case NEXTKEY:
        fk_set++;
	if (fk_set == NUMFK)
	  fk_set = 0;
	print_fk(fk_set);
	break;

      case PREVKEY:
        fk_set--;
	if (fk_set < 0)
	  fk_set = NUMFK - 1;
	print_fk(fk_set);
	break;
	
      case ROLLDOWN:
        pop(&tempx, &tempxf);
	t = tempx;
	tf = tempxf;
	dispnums();
	break;

      case CLX: /* Clear X */
        pop(&l, &lf);
	tempx = 0;
	tempxf = 0.0;
	push(tempx, tempxf);
	lift_needed = FALSE;
	dispnums();
	break;

      case ASCII:
	if (base != 1)
	  {
	  stop_entering();
	  base = 1;
	  msg("Only function keys work in ASCII mode");
	  max_digits = 4;
          dispnums();
	  dispregs();
	  prinbase();
	  }
	break;

      case FLOAT:
        if (((tempx = get_int("Enter number of decimal places")) >= 0) && (tempx != -base))
          {
          stop_entering();
          base = -tempx;
	  max_digits = 32;
          dispnums();
	  dispregs();
	  prinbase();
          }
        break;

      case ENG:
        if (((tempx = get_int("Enter number of decimal places")) >= 0) && (tempx != -(10 + base)))
          {
          stop_entering();
          base = -(10 + tempx);
	  max_digits = 32;
          dispnums();
	  dispregs();
	  prinbase();
          }
        break;

      case BIN:
	if (base != 2)
	  {
          stop_entering();
	  max_digits = 32;
          base = 2;
          dispnums();
	  dispregs();
	  prinbase();
	  }
        break;
	
      case OCT:
	if (base != 8)
	  {
          stop_entering();
          base = 8;
	  max_digits = 11;
          dispnums();
	  dispregs();
	  prinbase();
	  }
        break;

      case DEC:
	if (base != 10)
	  {
          stop_entering();
          base = 10;
	  max_digits = 10;
          dispnums();
	  dispregs();
	  prinbase();
	  }
        break;

      case HEX:
	if (base != 16)
	  {
          stop_entering();
          base = 16;
	  max_digits = 8;
          dispnums();
	  dispregs();
	  prinbase();
	  }
        break;

      case BACKSPACE:
      case '0':
      case '1':
        echo_char(*c);
        break;

      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
        if (base == 2)
          put_a_char(BELL);
        else
          echo_char(*c);
        break;

      case '8':
      case '9':
        if ((base == 2) || (base == 8))
          put_a_char(BELL);
        else
          echo_char(*c);
        break;

      case 'A':
      case 'B':
      case 'C':
      case 'D':
      case 'F':
        if (base != 16)
          put_a_char(BELL);
        else
          echo_char(*c);
        break;
 
      case 'E':
        if (base == 16)
	  echo_char(*c);
	else if (base > 0)
	  put_a_char(BELL);
	else
	  {
	  if (e_legal)
	    {
	    echo_char(*c);
	    e_legal = FALSE;
	    }
	  else
	    put_a_char(BELL);
	  }
	break;
	
      case '.':
        if (base > 0) /* base == 1 (ASCII) never reaches here! */
          put_a_char(BELL);
        else if (entering && !dot_legal)
	  put_a_char(BELL);
	else
          echo_char(*c);
        dot_legal = FALSE;
        break;
 
      case REFRESH: /* ^L */
        display();
        break;
 
      case '+':
      case '-':
      case '*':
      case '/':
      case '%':
      case '^':
      case '|':
      case '&':
      case '>':
      case '<':
      case '#':
	save_x();
        pop(&tempx, &tempxf);
        pop(&tempy, &tempyf);
	switch (*c)
	  {
	  case '+':
            push((long) (tempx + tempy), (double) (tempxf + tempyf));
	    break;
	  case '-':
            push((long) (tempy - tempx), (double) (tempyf - tempxf));
	    break;
	  case '*':
            push((long) (tempx * tempy), (double) (tempxf * tempyf));
	    break;
	  case '/':
	    if (((base > 0) && (tempx == 0)) || ((base < 0) && (tempxf == 0.0)))
	      {
	      push(tempy, tempyf);
	      push(tempx, tempxf);
	      msg("Divide by 0!\007");
	      }
	    else
	      {
	      if (tempx != 0)
	        tempx = tempy / tempx;
	      else
	        tempx = 0;
              push(tempx, tempyf / tempxf);
	      }
	    break;
	  case '%':
	    push(tempx * tempy / 100, tempxf * tempyf / 100.0);
	    break;
	  case '^':
	    tempx ^= tempy;
	    tempxf = tempx;
	    push(tempx, tempxf);
	    break;
	  case '|':
	    tempx |= tempy;
	    tempxf = tempx;
	    push(tempx, tempxf);
	    break;
	  case '&':
	    tempx &= tempy;
	    tempxf = tempx;
	    push(tempx, tempxf);
	    break;
	  case '<':
	    tempy <<= tempx;
	    tempyf = tempy;
	    push(tempy, tempyf);
	    break;
	  case '>':
	    tempy >>= tempx;
	    tempyf = tempy;
	    push(tempy, tempyf);
	    break;
	  case '#':
	    tempy %= tempx;
	    tempyf = fmod(tempyf, tempxf);
	    push(tempy, tempyf);
	    break;
	  }
        dispnums();
        break;
 
      case '~':
	save_x();
        pop(&tempx, &tempxf);
	tempx = ~tempx;
	tempxf = tempx;
	push(tempx, tempxf);
	dispnums();
	break;

      case '\012': /* Line feed */
      case '\015': /* Carriage return (Enter on PC) */
        pop(&tempx, &tempxf);
        push(tempx, tempxf);
        push(tempx, tempxf);
	lift_needed = FALSE;
        dispnums();
        break;

      case 'L': /* Last X */
	push(l, lf);
	dispnums();
        break;
	 
      case QUIT:
	msg("Really quit? (Y or N)");
	*c = get_a_char(&d);
	if (toupper(*c) == 'Y')
	  *c = QUIT;
	else
	  *c = '.';
        clear_msg();
        break;
 
      case 'R':
	i = get_int("Recall from register [0-9]");
	if ((i >= 0) && (i < NUMREGS))
	  {
	  tempx = reg[i];
	  tempxf = regf[i];
	  push(tempx, tempxf);
	  dispnums();
	  }
	break;

      case 'S':
        stop_entering();
	i = get_int("Store in register [0-9]");
	if ((i >= 0) && (i < NUMREGS))
	  {
	  reg[i] = x;
	  regf[i] = xf;
	  dispreg(i);
	  }
	break;

      case 'X':
        pop(&tempx, &tempxf);
	pop(&tempy, &tempyf);
	push(tempx, tempxf);
	push(tempy, tempyf);
	dispnums();
	break;

      case 'Y':
        save_x();
	pop(&tempx, &tempxf);
	pop(&tempy, &tempyf);
	
	/* if x < 0 then make sure y is integer ... */
	if (tempxf < 0.0)
	  {
	  tempx = 0;
	  if (tempyf < 0.0)
	    {
	    tempx = 1;
	    tempyf = -tempyf;
	    }
	  tempy = floor(tempyf + 0.5);
	  tempyf = tempy;
	  if (tempx)
	    tempyf = -tempyf;
	  }
	  
	tempxf = pow(tempyf, tempxf);
	tempx = tempxf;
	push(tempx, tempxf);
	dispnums();
	break;

      case ',': /* CHS */
        if (entering && (base < 0) && !e_legal) /* then CHS of exponent */
	  {
	  c_ptr = inbuf;
	  while (*c_ptr != 'E')
	    c_ptr++;
	  c_ptr++;
	  strcpy(c_buf, c_ptr);
	  *c_ptr = ASCIIZ;
	  (*c_buf == '-') ? strcat(inbuf, "+"): strcat(inbuf, "-");
	  if ((*c_buf == '-') || (*c_buf == '+'))
	    strcat(inbuf, &c_buf[1]);
	  else
	    strcat(inbuf, c_buf);
	  inptr = &inbuf[strlen(inbuf)];
	  }
	else
	  {
          pop(&tempx, &tempxf);
	  push(-tempx, -tempxf);
	  }
	dispnums();
	break;

/* TRIG FUNCTIONS ... */

      case INV:
        invert = !invert;
	print_inv();
	break;
	
      case 'M':
        degree = !degree;
	print_deg();
	break;
	
      case SIN:
	save_x();
        pop(&tempx, &tempxf);
        if (invert)
	  {
	  if (degree)
	    tempxf = r_to_d * asin(tempxf);
	  else
	    tempxf = asin(tempxf);
	  invert = FALSE;
	  print_inv();
	  }
	else
	  {
	  if (degree)
	    tempxf = sin(d_to_r * tempxf);
	  else
	    tempxf = sin(tempxf);
	  }
        tempx = tempxf;
	push(tempx, tempxf);
	dispnums();
	break;
	
      case COS:
	save_x();
        pop(&tempx, &tempxf);
        if (invert)
	  {
	  if (degree)
	    tempxf = r_to_d * acos(tempxf);
	  else
	    tempxf = acos(tempxf);
	  invert = FALSE;
	  print_inv();
	  }
	else
	  {
	  if (degree)
	    tempxf = cos(d_to_r * tempxf);
	  else
	    tempxf = cos(tempxf);
	  }
        tempx = tempxf;
	push(tempx, tempxf);
	dispnums();
	break;
	
      case TAN:
	save_x();
        pop(&tempx, &tempxf);
        if (invert)
	  {
	  if (degree)
	    tempxf = r_to_d * atan(tempxf);
	  else
	    tempxf = atan(tempxf);
	  invert = FALSE;
	  print_inv();
	  }
	else
	  {
	  if (degree)
	    tempxf = tan(d_to_r * tempxf);
	  else
	    tempxf = tan(tempxf);
	  }
        tempx = tempxf;
	push(tempx, tempxf);
	dispnums();
	break;

      case LOGE:
	save_x();
        pop(&tempx, &tempxf);
	if (invert)
	  {
	  tempxf = exp(tempxf);
	  tempx = tempxf;
	  invert = FALSE;
	  print_inv();
	  }
	else
	  {
	  tempxf = log(tempxf);
	  tempx = tempxf;
	  }
	push(tempx, tempxf);
	dispnums();
	break;
		
      case LOG10:
	save_x();
        pop(&tempx, &tempxf);
	if (invert)
	  {
	  tempxf = pow(10.0, tempxf);
	  tempx = tempxf;
	  invert = FALSE;
	  print_inv();
	  }
	else
	  {
	  tempxf = log10(tempxf);
	  tempx = tempxf;
	  }
	push(tempx, tempxf);
	dispnums();
	break;
		
      case 'P':
        tempxf = pi;
	tempx = tempxf;
	push(tempx, tempxf);
	dispnums();
	break;

/* Square root keys ... */

      case SQR:
	save_x();
        pop(&tempx, &tempxf);
	push(tempx * tempx, tempxf * tempxf);
	dispnums();
	break;
	
      case ROOT:
	save_x();
        pop(&tempx, &tempxf);
	tempyf = tempx;
	tempx = round(sqrt(tempyf));
	tempxf = sqrt(tempxf);
	push(tempx, tempxf);
	dispnums();
	break;
	
      case CUBE:
	save_x();
        pop(&tempx, &tempxf);
	push(tempx * tempx * tempx, tempxf * tempxf * tempxf);
	dispnums();
	break;
	
      case CROOT:
	save_x();
        pop(&tempx, &tempxf);
	tempyf = tempx;
	tempx = round(pow(tempyf, 1.0 / 3.0));
	tempxf = pow(tempxf, 1.0 / 3.0);
	push(tempx, tempxf);
	dispnums();
	break;
	
      case FRC:
	save_x();
        pop(&tempx, &tempxf);
	tempy = 0;
	if (tempxf < 0.0)
	  {
	  tempy = 1;
	  tempxf = -tempxf;
	  }
	tempxf = tempxf - floor(tempxf);
	if (tempy)
	  tempxf = -tempxf;
	push(0L, tempxf);
	dispnums();
	break;
	
      case INT:
	save_x();
        pop(&tempx, &tempxf);
	tempy = 0;
	if (tempxf < 0.0)
	  {
	  tempy = 1;
	  tempxf = -tempxf;
	  }
	tempxf = (double) floor(tempxf);
	if (tempy)
	  tempxf = -tempxf;
	push(tempx, tempxf);
	dispnums();
	break;

      case SUM:
	save_x();
	pop(&tempx, &tempxf);
	reg[7]++;
	regf[7]++;
	reg[8] += tempx;
	regf[8] += tempxf;
	reg[9] += tempx * tempx;
	regf[9] += tempxf * tempxf;
	push(reg[7], regf[7]);
	dispreg(7);
	dispreg(8);
	dispreg(9);
	dispnums();
	msg("X accumulated");
	break;
	
      case SUMR:
	if (reg[7] > 0)
	  {
	  save_x();
	  pop(&tempx, &tempxf);
	  reg[7]--;
	  regf[7]--;
	  reg[8] -= tempx;
	  regf[8] -= tempxf;
	  reg[9] -= tempx * tempx;
	  regf[9] -= tempxf * tempxf;
	  dispreg(7);
	  dispreg(8);
	  dispreg(9);
	  dispnums();
	  msg("X decremented");
	  }
	break;
	
      case SUM0:
        reg[7] = 0;
	regf[7] = 0.0;
	reg[8] = 0;
	regf[8] = 0.0;
	reg[9] = 0;
	regf[9] = 0.0;
	dispreg(7);
	dispreg(8);
	dispreg(9);
	dispnums();
	msg("Accumulator registers cleared");
	break;

      case NSTAT:
        push(reg[7], regf[7]);
	dispnums();
	break;
		
      case MEAN:
	tempxf = regf[8] / regf[7];
	tempx = tempxf;
	push(tempx, tempxf);
	dispnums();
	break;
	
      case S_DEV:
        /* s.dev = sqrt( n * sum(x^2) - (sum(x))^2 ) / n */
        tempxf = sqrt( regf[7] * regf[9] - regf[8] * regf[8]) / regf[7];
	tempx = tempxf;
	push(tempx, tempxf);
	dispnums();
	break;

/*
 * Financial function register allocation ...
 *
 * 5 n
 * 6 interest
 * 7 present value
 * 8 payment
 * 9 future value
 *
 * Formulae are :
 *
 * 0 = PVAL + (1 + I.S) PMT (1 - (1+I)**-N ) / I + FVAL * (1+I)**-N
 *
 * where:
 *
 * N    = number of periods
 * PVAL = present value
 * I    = interest rate (as a fraction - user sees a percent)
 * PMT  = payment
 * FVAL = future value
 * S    = payment mode factor. 0 for payment at end of period, 1 for start.
 *        This version only uses S=0.
 */
      case PVAL:
	if (!last_was_fin_key)
	  {
	  save_x();
	  pop(&tempx, &tempxf);
	  regf[7] = tempxf;
	  reg[7] = 0;
	  dispreg(7);
	  dispnums();
	  this_was_fin_key = TRUE;
	  msg("Present value entered");
	  }
	else
	  {
	  n = regf[5];
	  interest = regf[6] / 100.0;
	  pmt = regf[8];
	  fval = regf[9];
	  tempyf = pow(1.0 + interest, -n);
	  tempxf = pmt * (1.0 - tempyf) / interest
			+ fval * tempyf;
	  tempxf = -tempxf;
	  push(0L, tempxf);
	  regf[7] = tempxf;
	  reg[7] = 0;
	  dispreg(7);
	  dispnums();
	  msg("Present value computed");
	  }
	break;

      case FVAL:
	if (!last_was_fin_key)
	  {
	  save_x();
	  pop(&tempx, &tempxf);
	  regf[9] = tempxf;
	  reg[9] = 0;
	  dispreg(9);
	  dispnums();
	  this_was_fin_key = TRUE;
	  msg("Final value entered");
	  }
	else
	  {
	  n = regf[5];
	  interest = regf[6] / 100.0;
	  pval = regf[7];
	  pmt = regf[8];
	  tempyf = pow(1.0 + interest, n);
	  tempxf = (pval + pmt * (1.0 - 1.0 / tempyf) / interest) * tempyf;
	  tempxf = -tempxf;
	  push(0L, tempxf);
	  regf[9] = tempxf;
	  reg[9] = 0;
	  dispreg(9);
	  dispnums();
	  msg("Final value calculated");
	  }
	break;

      case PMT:
	if (!last_was_fin_key)
	  {
	  save_x();
	  pop(&tempx, &tempxf);
	  regf[8] = tempxf;
	  reg[8] = 0;
	  dispreg(8);
	  dispnums();
	  this_was_fin_key = TRUE;
	  msg("Payment entered");
	  }
	else
	  {
	  n = regf[5];
	  interest = regf[6] / 100.0;
	  pval = regf[7];
	  fval = regf[9];
	  tempyf = pow(1.0 + interest, -n);
	  tempxf = (pval + fval * tempyf) * interest / (1.0 - tempyf);
	  tempxf = -tempxf;
	  push(0L, tempxf);
	  regf[8] = tempxf;
	  reg[8] = 0;
	  dispreg(8);
	  dispnums();
	  msg("Payment calculated");
	  }
	break;

      case INTST:
	if (!last_was_fin_key)
	  {
	  save_x();
	  pop(&tempx, &tempxf);
	  regf[6] = tempxf;
	  reg[6] = 0;
	  dispreg(6);
	  dispnums();
	  this_was_fin_key = TRUE;
	  msg("Interest rate (%) entered");
	  }
	else
	  {
	  n = regf[5];
	  pval = regf[7];
	  pmt = regf[8];
	  fval = regf[9];
	  /* use Newton-Raphson iteration ...

	    if we are seeking roots of f(x)=0 then

	    x_n+1 = x_n - f(x_n) / f'(x_n)
	  */
	  count = 0;
	  epsilon = 1000.0;
	  guess = 0.01; /* == 12% per annum */
	  while ((count < 10000) && (fabs(epsilon) > 0.0001))
	    {
	    tempyf = pow(1.0 + guess, -n);
	    epsilon = ((pmt + pval * guess) - (pmt - fval * guess) * tempyf)
	    / (n * (pmt + pval * guess) / (1.0 + guess) + pval + fval * tempyf);
	    guess -= epsilon;
	    count++;
	    }
	  tempxf = guess * 100.0; /* Convert to percent */
	  if (count < 10000)
	    {
	    push(0L, tempxf);
	    regf[6] = tempxf;
	    reg[6] = 0;
	    dispreg(6);
	    dispnums();
	    msg("Interest rate (%) computed");
	    }
	  else
	    msg("Calculation does not converge!\007");
	  }
	break;

      case NFIN:
	if (!last_was_fin_key)
	  {
	  save_x();
	  pop(&tempx, &tempxf);
	  regf[5] = tempxf;
	  reg[5] = 0;
	  dispreg(5);
	  dispnums();
	  this_was_fin_key = TRUE;
	  msg("Number of payments entered");
	  }
	else
	  {
	  interest = regf[6] / 100.0;
	  pval = regf[7];
	  pmt = regf[8];
	  fval = regf[9];
	  tempxf = log((pmt - fval * interest) / (pmt + pval * interest))
	      	 / log(1.0 + interest);
	  push(0L, tempxf);
	  regf[5] = tempxf;
	  reg[5] = 0;
	  dispreg(5);
	  dispnums();
	  msg("Number of payments computed");
	  }
	break;

      default:
        put_a_char(BELL);
        break;
      }
  last_was_fin_key = FALSE;
  if (this_was_fin_key)
    last_was_fin_key = TRUE;
}

#ifndef HPUNIX
CXFERR(err)
int err;

{
  msg("Floating point error!\007");
  longjmp(&save, 0);
}
#endif

matherr(x)
struct exception *x;

{
  msg("Math error!\007");
  longjmp(&save, 0);
}

#ifdef HPUNIX
prep_signal()

{
  int do_signal();
  int terminate();
  int i;

  for (i = 0; i < NSIG; i++)
    signal(i, terminate);
  signal(SIGFPE, do_signal);
}

do_signal()

{
  prep_signal();
  msg("Floating point error!\007");
  longjmp(&save, 0);
}
#endif 

clear_msg()

{
  int i;

  if (msg_flag) /* set by msg() */
    {
    gotoxy(ERRORX, ERRORY);
    *outbuf = '\0';
    while (strlen(outbuf) < 40)
      strcat(outbuf, " ");
    print_attr(outbuf, 0);
    msg_flag = 0;
    }
}

initialise()

{
  int i;

#ifdef HPTERM
  print_string("\033&s1A\033&s1g1H");
  print_string("\033&f0a1k0d2L\033p");
  print_string("\033&f0a2k0d2L\033q");
  print_string("\033&f0a3k0d2L\033r");
  print_string("\033&f0a4k0d2L\033s");
  print_string("\033&f0a5k0d2L\033t");
  print_string("\033&f0a6k0d2L\033u");
  print_string("\033&f0a7k0d2L\033v");
  print_string("\033&f0a8k0d2L\033w");
  print_string("\033&jB");
#endif

#ifdef HPUNIX
  uncanon();
#endif

  x = y = z = t = l = 0;
  xf = yf = zf = tf = lf = 0.0;
  base = -2;
  dot_legal = TRUE;
  e_legal = TRUE;
  entering = FALSE;
  lift_needed = TRUE;
  invert = FALSE;
  degree = FALSE;
  msg_flag = FALSE;
  last_was_fin_key = FALSE;
  inptr = inbuf;
  *inptr = ASCIIZ;
  fk_set = 0;
  max_digits = 32;
  pi = 4.0 * atan(1.0);
  d_to_r = pi / 180.0;
  r_to_d = 180.0 / pi;
  
  for (i = 0; i < NUMREGS; i++)
    {
    reg[i] = 0;
    regf[i] = 0.0;
    }

#ifdef HPUNIX
  prep_signal();
#endif
 
  display();
#ifndef HPUNIX
  msg("DCALC: (C)Copyright Prism Software 1986");
#endif
}

terminate()

{
  gotoxy(0, 23);

#ifdef HPUNIX
  canon();
#endif

#ifdef HPTERM
  print_string("\033&s0A");
#endif
  exit(0);
}

void print_base(x, xf)
    long	x;
    double	xf;
{
    fmt_base(outbuf, x, xf);
    prep_for_output(outbuf);
    print_string(outbuf);
}
  
void dispreg(i)
    int	i;
{
    if (reg_select()) {
	gotoxy(LEFTSIDE, i);
	outbuf[0] = '0' + i;
	outbuf[1] = '\0';
	print_string(outbuf);
	gotoxy(START_X, i);
	print_base(reg[i], regf[i]);
    }
}

void dispregs()
{
    int i;
    
    for (i = 0; i < NUMREGS; i++)
	dispreg(i);
}

void prinbase()
{
    xwin_select();
    gotoxy(BASE_X, BASE_Y);
    switch (base) {
    case 1:
	print_string("Ascii   ");
	break;
	
    case 2:
	print_string("Binary  ");
	break;
	
    case 8:
	print_string("Octal   ");
	break;
	
    case 10:
	print_string("Decimal ");
	break;
	
    case 16:
	print_string("Hex     ");
	break;
	
    default:
	if (base < -9) {
	    sprintf(outbuf, "Engin. %d", -(base + 10));
	    print_string(outbuf);
	} else {
	    sprintf(outbuf, "Float  %d", -base);
	    print_string(outbuf);
	}
	break;
    }
}

void print_inv()
{
    xwin_select();
    gotoxy(INV_X, INV_Y);
    if (invert)
	print_string("Inv   ");
    else
	print_string("      ");
}

void print_deg()
{
    xwin_select();
    gotoxy(TMODE_X, TMODE_Y);
    if (degree)
	print_string("Degrees");
    else
	print_string("Radians");
}

void print_x(char *buf) 
{
    gotoxy(LEFTSIDE, STARTX_Y);
    print_string(buf);
}

void dispnums()
{
    xwin_select();
    gotoxy(LEFTSIDE, STARTX_Y);
    if (entering) {
	strcpy(outbuf, inbuf);
	prep_for_output(outbuf);
	print_string(outbuf);
    } else
	print_base(x, xf);
    
    if (reg_select()) {
	gotoxy(LEFTSIDE, STARTT_Y);
	print_string("T");
	gotoxy(START_X, STARTT_Y);
	print_base(t, tf);
	
	gotoxy(LEFTSIDE, STARTZ_Y);
	print_string("Z");
	gotoxy(START_X, STARTZ_Y);
	print_base(z, zf);
	
	gotoxy(LEFTSIDE, STARTY_Y);
	print_string("Y");
	gotoxy(START_X, STARTY_Y);
	print_base(y, yf);
	
	gotoxy(LEFTSIDE, STARTL_Y);
	print_string("L");
	gotoxy(START_X, STARTL_Y);
	print_base(l, lf);
    }
}


main(argc, argv)
int argc;
char **argv;

{
  char c, d;
 
  line_drawing = TRUE;
#ifdef HPUNIX
  if (argc > 1)
    line_drawing = FALSE;
#endif

  initialise();
  setjmp(&save); 
  do 
    {
    c = get_a_char(&d);
    clear_msg();
    c = map_key_to_func(c, d);
    process(c);
    setjmp(&save);
    }
  while (c != QUIT);

  terminate();
}

