#include <bvideo.h>
#include <dos.h>
#include "dcalc.h"

#define ESC	'\033'
#define CR 	'\r'
#define PLUS	'+'
#define MINUS	'-'
#define DIVIDE	'/'
#define MULT	'*'

static int cursor = 0, region = 0, x1old, y1old, x2old, y2old;
extern char get_a_char();

void invert_char(int x, int y)

{
  char b[2];
  int fore, back;

  virdrect(y, x, y, x, b, CHAR_ATTR);
  fore = ((b[1] >> 4) & 7) | (b[1] & 8);
  back = (b[1] & 7) | ((b[1] >> 4) & 8);
  viatrect(y, x, y, x, fore, back);
}

void cursor_on(int x, int y)

{
  if (cursor && (x == x1old) && (y == y1old))
    return;

  if (cursor)
    invert_char(x1old, y1old);
  invert_char(x, y);
  x2old = x1old = x;
  y2old = y1old = y;
  cursor = 1;
}

void cursor_off(void)

{
  if (cursor)
    {
    invert_char(x1old, y1old);
    cursor = 0;
    }
}

void region_on(int x1, int y1, int x2, int y2)

{
  int x, y, t, minx1, maxx1, miny1, maxy1, minx2, maxx2, miny2, maxy2;

  if (x1 > x2)
    {
    t = x1;
    x1 = x2;
    x2 = t;
    }
  if (y1 > y2)
    {
    t = y1;
    y1 = y2;
    y2 = t;
    }

  if (region && (x1 == x1old) && (y1 == y1old) &&
  		(x2 == x2old) && (y2 == y2old))
    return;

  /* Undraw old except where it is inside new region */
  if (region)
    for (y = y1old; y <= y2old; y++)
      for (x = x1old; x <= x2old; x++)
        if ((x >= x1) && (x <= x2) && (y >= y1) && (y <= y2))
	  ;
        else
          invert_char(x, y);

  /* Draw new except where it is inside old region */
  for (y = y1; y <= y2; y++)
    for (x = x1; x <= x2; x++)
      if (region && (x >= x1old) && (x <= x2old) && (y >= y1old) && (y <= y2old))
	;
      else
        invert_char(x, y);

  x1old = x1;
  y1old = y1;
  x2old = x2;
  y2old = y2;
  region = 1;
}

void region_off(void)

{
  int x, y;

  if (region)
    for (y = y1old; y <= y2old; y++)
      for (x = x1old; x <= x2old; x++)
        invert_char(x, y);

  region = 0;
}

int get_mouse(int *i, int *j)

{
  union REGS inregs, outregs;

  inregs.x.ax = 3;
  int86(51, &inregs, &outregs); /* Get mouse position & button status */
  *i = outregs.x.cx >> 3;
  *j = outregs.x.dx >> 3;
  return(outregs.x.bx);
}

int echo_arrow(int *i, int *j)

{
  char c, d;

  c = get_a_char(&d);
  if (c == 0)
    {
    switch(d)
      {
      case UPKEY  :  (*j)--;		 break;
      case DOWNKEY:  (*j)++;		 break;
      case LEFTKEY:  (*i)--;		 break;
      case RIGHTKEY: (*i)++; 		 break;
      case HOMEKEY:  (*i) = 0;  *j = 0;  break;
      case PREVKEY:  (*i) = 79; *j = 0;  break;
      case NEXTKEY:  (*i) = 79; *j = 24; break;
      case ENDKEY:   (*i) = 0;  *j = 24; break;
      }
    if (*j > 24) *j = 24;
    if (*j <  0) *j = 0;
    if (*i > 79) *i = 79;
    if (*i <  0) *i = 0;
    }
  else
    if ((c != ESC) && (c != CR))
      c = 0;
  return(c);
}

int get_region(int *x1, int *y1, int *x2, int *y2)

{
  char c;
  union REGS inregs, outregs;
  int mouse_present = 0;
  int button, d;

  button = 0;

                          /* Detect mouse */
  inregs.x.ax = 0;
  int86(51, &inregs, &outregs);
  if (outregs.x.ax != 0)
    {
    /* num_buttons = outregs.x.bx; .... in case I ever need it */
    mouse_present = 1;
    /*
    inregs.x.ax = 1;
    int86(51, &inregs, &outregs); /* Turn on sprite */

    inregs.x.ax = 4;
    inregs.x.cx = *x1;
    inregs.x.dx = *y1;
    int86(51, &inregs, &outregs); /* set cursor position */

    /* wait for buttons up */
    do
      button = get_mouse(&d, &d);
    while (button != 0);
    }

do
  {
			/* Get first point */
  button = 0;
  while (button == 0)
    {
    if (mouse_present)
      button = get_mouse(x1, y1);

    if (kbhit())
      button = echo_arrow(x1, y1);

    cursor_on(*x1, *y1);
    }

  cursor_off();

			/* Get second point */
  if ((button == 1) || (button == CR))
    {
    *x2 = *x1;
    *y2 = *y1;
    region_on(*x1, *y1, *x2, *y2);

    button = 0;
    while (button == 0)
      {
      if (mouse_present)
	button = get_mouse(x2, y2) ? 0 : 1; /* ie wait for button release */

      if (kbhit())
        button = echo_arrow(x2, y2);
      region_on(*x1, *y1, *x2, *y2);
      }
    }

/* If using mouse, repeat until carriage return/left button or ESC/right button */
  if (mouse_present && (button != ESC) && (button != 2))
    {
    button = 0;
    while (button == 0)
      {
      if (kbhit())
        button = get_a_char(&c);
      else
        button = get_mouse(&d, &d);
      }
    }
  region_off();

  }
while ((button != 1) &&
       (button != CR) &&
       (button != PLUS) &&
       (button != MINUS) &&
       (button != DIVIDE) &&
       (button != MULT) &&
       (button != ESC));

  return(button);
}
