/* Platform independent parts of DCALC - Reverse Polish Notation (RPN)
 * calculator.

    Contact info:
    bhepple@bit.net.au
    http://www.bit.net.au/~bhepple/dcalc

    Copyright (C) 1999,2000 Bob Hepple

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.

 */

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h> /* Needed for Linux because atof() is not declared in math.h! */
#include <unistd.h> /* for pipe() */
#include <sys/types.h> /* for wait */
#include <sys/wait.h> /* for wait */
#include <time.h> /* for localtime */

#ifdef HAVE_CONFIG_H /* gtk */
#  include <config.h>
#endif

#include "dcalc.h"

long    x, y, z, t, l,  /* Integer stack                             */
    reg[NUMREGS];   /* Integer registers                         */
/*                                           */
double  xf,yf,zf,tf,lf, /* Floating point stack                      */
    regf[NUMREGS],  /* Floating point registers                  */
    pi,             /* The usual pi.                             */
    d_to_r,         /* Degrees to radians factor                 */
    r_to_d;         /* Radians to degrees factor                 */
/*                                           */
BOOLEAN entering,       /* True if we are still entering into X      */
    lift_needed,    /* True if the stack needs lifting before    */
    /* more characters are put into X            */
    invert,         /* Invert flag - causes the next operation   */
    /* to be inversed eg arcsin() instead of sin */
    degree,         /* True if numbers are in degrees; else      */
    /* radians                                   */
    was_initialised = 0;
/* To make sure program is only initialised once */
static BOOLEAN
last_was_fin_key;/* True if last operation was financial      */
BOOLEAN finPayAt0 = 0;

char
inbuf[45],      /* Where typed characters are put            */
    *inptr,         /* Pointer into inbuf                        */
    outbuf[45],     /* General buffer used to format output      */
    zero_string[] = "0",
    div_by_zero[] = "Divide by 0!";

int
decplaces,	/* number of decimal places		     */
    max_digits;     /* Maximum number of digits displayable      */

int	mode, intmode, floatmode;

typedef unsigned char byte;

/* function prototypes: */
static long asc2int(char *s);
static long bin2int(char *s);
static void stop_entering(void);
static void push(long inx);
static void pushf(double inx);
static long pop();
static double popf();
static void echo_char(byte c);
static void save_x(void );
#ifndef __hpux
static long round(double f);
#endif
static void change_base(COMMAND c);
static void arith_int(COMMAND c);
static void arith_real(COMMAND c);
static void trig(COMMAND c);
static void summation(COMMAND c);
static void financial(COMMAND c);
static void process_digit(COMMAND c);
static int isInverted();
static long ip2int(char *s);

static int isInverted() {
    int retval = invert;

    invert = FALSE;
    print_inv();
    clear_msg();
    return retval;
}

static void fmt_ip(char *str, int x) {
    int negative, i1,i2,i3,i4;

    if (x < 0) {
	x = x & 0x7ffffff;
	negative = 1;
    } else {
	negative = 0;
    }
    i4 = x & 0xff;
    x = x >> 8;
    i3 = x & 0xff;
    x = x >> 8;
    i2 = x & 0xff;
    x = x >> 8;
    if (negative)
	x = x | 0x80;
    i1 = x & 0xff;
    sprintf(str, "%03d.%03d.%03d.%03d", i1, i2, i3, i4);
}

void fmt_bin(char *str, long x) {
    char 	inverted[50],
	reverted[50],
	*s, *t;
    BOOLEAN negative;
    
    if (x < 0) {
	negative = TRUE;
	x &= 0x7FFFFFFF;
    } else
	negative = FALSE;
    s = inverted;
    while (x != 0) {
	*(s++) = '0' + (char) (x % 2);
	x /= 2;
    }
    *s = ASCIIZ;
    if (negative) {
	while (strlen(inverted) < 31) {
	    *(s++) = '0';
	    *s = ASCIIZ;
	}
	*(s++) = '1';
    }
    if (s == inverted) /* x==0 */
	strcpy(str, zero_string);
    else {
	t = reverted;
	while (s != inverted)
	    *(t++) = *(--s);
	*t = ASCIIZ;
	sprintf(str, "%32.32s", reverted);
    }
}

void fmt_base(char *s, long x, double xf) {
    byte 	format[20];
    byte 	*p;
    union 	frig {
	long j;
	byte c[4];
    } frigger;
    int 	i, leading_zero;
    
    if (mode == PROG) {
	switch (intmode) {
	case ASCIIM:
	    frigger.j = x;
	    for (i = 0; i < 4; i++) {
#if REVERSE_BYTES
		format[3 - i] = frigger.c[i];
#else
		format[i] = frigger.c[i];
#endif
	    }
	    format[4] = 0;
#if PCSPECIFIC
	    p = format;
	    for (i = 0; i < 4; i++, p++)
		if (*p == 0)
		    *s++ = ' ';
		else
		    *s++ = *p;
	    *s = 0;
#else
	    /* Convert to canonical form */
	    p = format;
	    leading_zero = 1;
	    for (i = 0; i < 4; i++, p++) {
		if ((i < 3) && (*p == 0) && leading_zero)
		    continue;
		leading_zero = 0;
		if (((*p > 126) && (*p < 158)) || (*p < 32)) {
		    *s++ = '^';
		    *s++ = *p + '@';
		} else
		    *s++ = *p;
	    }
	    *s = 0;
#endif
	    break;
	
	case BIN:
	    fmt_bin(s, x);
	    break;
	
	case IP:
	    fmt_ip(s, x);
	    break;

	case OCT:
	    if (x == 0L)
		strcpy(s, zero_string);
	    else
		sprintf(s, "%32.lo", x);
	    break;
	case DEC:
	    if (x == 0L)
		strcpy(s, zero_string);
	    else
		sprintf(s, "%32.ld", x);
	    break;
	case HEX:
	    if (x == 0L)
		strcpy(s, zero_string);
	    else
		sprintf(s, "%32.lx", x);
	    break;
	}
    } else {
	double lowerLimit = pow(10.0, -decplaces);

	sprintf(format, "%%32.%d", decplaces);
	if (floatmode == ENG) 
	    strcat(format, "E");
	else if ((fabs(xf) > 1.0E25) || (fabs(xf) < lowerLimit))
	    strcat(format, "g");
	else
	    strcat(format, "f");
	sprintf(s, format, xf);
    }
}

void prep_for_output(char *string) {
    /* Put suitable separators in string */
    char 	*s, *d, sep = 0, *ft, rhs[45], lhs[45], scratch[45];
    int 	len, interval = 0, count, negative;
    
    if (mode == PROG) {
	switch (intmode) {
	case ASCIIM:
	    return; /* Nasty to return here - its easy to miss it when reading! */
	case IP:
	    return; /* all done by fmt_ip */
	case BIN:
	    sep = SPACE;
	    interval = 8;
	    break;
	case OCT:
	    sep = SPACE;
	    interval = 3;
	    break;
	case DEC:
	    sep = COMMA;
	    interval = 3;
	    break;
	case HEX:
	    sep = SPACE;
	    interval = 2;
	    break;
	}
    } else { /* FLOAT or ENG */
	sep = COMMA;
	interval = 3;
    }
    
    /* Find first non-blank ... */
    s = string;
    while (*s && isspace(*s)) s++;
    if (*s == '-') {
	negative = TRUE;
	*s++ = SPACE;
    } else
	negative = FALSE;
    
    /* move rhs to safe keeping - fraction or exponent part */
    ft = NULL;
    *rhs = 0;
    if ((mode != PROG) && ((ft = strchr(s, '.')) ||
			   (ft = strchr(s, 'E')) ||
			   (ft = strchr(s, 'e')))) {
	strcpy(rhs, ft);
	*ft = 0;
    }
    
    /* add separators to the integer part ... */
    strcpy(lhs, s);
    len = strlen(s);
    if (len > interval) {
	count = 0;
	d = scratch;
	s = s + len - 1;
	while (--len >= 0) {
	    if (count++ == interval) {
		*d++ = sep;
		count = 1;
	    }
	    *d++ = *s--;
	}
	*d = 0;
	s = scratch + strlen(scratch) - 1;
	d = lhs;
	while (s >= scratch)
	    *d++ = *s--;
	*d = 0;
    }
    
    /* Now reassemble ... */
    d = string;
    if (negative) {
	*d++='-';
    }
    *d = 0;
    strcpy(d, lhs);
    strcat(d, rhs);
}

static long asc2int(char *s) {
    union 	frig {
	long j;
	byte c[4];
    } frigger;
    int 	i;
    
    for (i = 0; i < 4; i++)
#if REVERSE_BYTES
	frigger.c[3 - i] = *s++;
#else
    frigger.c[i] = *s++;
#endif
    return(frigger.j);
}

static long bin2int(char *s) {
    char 	*t;
    long 	accum;
    
    accum = 0;
    t = s;
    while (*t)
	accum = 2 * accum + (*(t++) - '0');
    return(accum);
}

static long ip2int(char *s) {
    int index, i[4], leftBitSet = 0;
    long retval;

    if (!s || !*s)
	return 0;

    for (index = 0; index < 4; index++) {
	char *dot = strchr(s, '.');
	i[index] = 0;
	if (dot)
	    *dot = 0;
	if (strlen(s)) {
	    sscanf(s, "%d", &(i[index]));
	    i[index] &= 0xff;
	    s += strlen(s);
	}
	if (dot) {
	    *dot = '.';
	    s = dot + 1;
	}
    }
    if (i[0] > 0x7f) {
	leftBitSet = 1;
	i[0] &= 0x7f;
    }
    retval = i[0];
    retval = (retval << 8) | i[1];
    retval = (retval << 8) | i[2];
    retval = (retval << 8) | i[3];
    if (leftBitSet)
	retval |= 0x80000000;
    return retval;
}

static void stop_entering() {
    
    if (entering) {
	entering = FALSE;
	if (mode == PROG) {
	    long temp = 0;
	    switch (intmode) {
	    case ASCIIM:
		temp = asc2int(inbuf);
		break;
	    case IP:
		temp = ip2int(inbuf);
		break;
	    case BIN:
		temp = bin2int(inbuf);
		break;
	    case OCT:
		sscanf(inbuf, "%lo", &temp);
		break;
	    case DEC:
		sscanf(inbuf, "%ld", &temp);
		break;
	    case HEX:
		sscanf(inbuf, "%lx", &temp);
		break;
	    }
	    x = temp;
	} else {
	    double tempf = atof(inbuf);
	    xf = tempf;
	}
	inptr = inbuf;
	*inptr = 0;
    }
}

static void push(long inx) {
    stop_entering();
    t = z;     
    z = y;     
    y = x;     
    x = inx;   
    lift_needed = TRUE;
}

static void pushf(double inxf) {
    stop_entering();
    tf = zf;
    zf = yf;
    yf = xf;
    xf = inxf;
    lift_needed = TRUE;
}

static long pop() {
    long retval;
    stop_entering();
    retval = x;    
    x = y;        
    y = z;        
    z = t;
    return retval;
}

static double popf() {
    double retval;
    stop_entering();
    retval = xf;    
    xf = yf;        
    yf = zf;        
    zf = tf;
    return retval;
}

static void echo_char(byte c) {
    /* echo 'c' back to the screen and log in the input buffer */
    long 	temp;
    double 	tempf;
    char 	*p;
    
    if ((c == BACKSPACE) && (mode != PROG || intmode != ASCIIM)) {
	if (entering && (inptr > inbuf)) {
	    inptr--;
	    *inptr = ASCIIZ;
	} else {
	    x = 0;
	    xf = 0.0;
	    inptr = inbuf;
	    *inptr = ASCIIZ;
	    entering = TRUE;
	    dispnums();
	}

    } else /* Neither backspace or ASCII mode */ {
	if (!entering) {
	    if (lift_needed) {
		if (mode == PROG) {
		    temp = x;
		    push(temp); 
		} else {
		    tempf = xf;
		    pushf(tempf);
		}
		dispnums();
	    }
	    lift_needed = TRUE;
	    entering = TRUE;
	    for (inptr = inbuf; inptr < inbuf + 4; inptr++)
		*inptr = ASCIIZ;
	    inptr = inbuf;
	}
	
	if (inptr < inbuf + max_digits)
	    if (mode == PROG && intmode == ASCIIM) {
		for (p = inbuf; p < inbuf + 3; p++)
		    *p = *(p+1);
		inbuf[3] = c;
		inptr++;
	    } else {
		*inptr++ = c;
		*inptr = ASCIIZ;
	    }
	else {
	    put_a_char(BELL);
	    msg("Too many digits!");
	}
    }
    
    if (mode != PROG || intmode != ASCIIM)
	strcpy(outbuf, inbuf);
    else {
	x = asc2int(inbuf);
	xf = x;
	fmt_base(outbuf, x, xf);
    }
    prep_for_output(outbuf);
    print_x(outbuf);
}

void display() {
    prinbase();
    print_deg();
    dispnums();
    dispregs();
    invert = FALSE;
    print_inv();
}

static void save_x() {
    stop_entering();
    l = x;
    lf = xf;
}

#ifndef __hpux
static long round(double f) {
    BOOLEAN	negative;
    long 	ret;
    
    negative = FALSE;
    if (f < 0.0) {
	negative = TRUE;
	f = -f;
    }
    ret = floor(f + 0.5);
    if (negative)
	ret = -ret;
    return(ret);
}
#endif

static void change_base(COMMAND c) {
    long	tempx;
    
    switch (c) {
    case ASCIIM:
	if (mode != PROG || intmode != ASCIIM) {
	    mode = PROG;
	    stop_entering();
	    intmode = ASCIIM;
	    msg("ASCII mode");
	    os_raw_mode(1);
	    max_digits = 4;
	}
	break;
	
    case ENG:
    case FLOAT:
	if (c != floatmode) {
	    char buf[80];

	    floatmode = c;
	    stop_entering();
	    os_raw_mode(0);
	    max_digits = 20;
	    sprintf(buf, "%s mode", (c==ENG)? "Engineering": "Floating point");
	    msg(buf);
	}
	break;
	
    case PLACES:
	if (((tempx = places("Enter number of decimal places")) >= 0) &&
	    ((tempx != decplaces) || (c != floatmode))) {
	    char buf[80];

	    stop_entering();
	    decplaces = tempx;
	    os_raw_mode(0);
	    max_digits = 20;
	    sprintf(buf, "Places set to %d", decplaces);
	    msg(buf);
	}
	break;
	
    case IP:
	if (mode != PROG || intmode != IP) {
	    mode = PROG;
	    stop_entering();
	    os_raw_mode(0);
	    max_digits = 15;
	    intmode = IP;
	    msg("Programming mode");
	}
	break;
	
    case BIN:
	if (mode != PROG || intmode != BIN) {
	    mode = PROG;
	    stop_entering();
	    os_raw_mode(0);
	    max_digits = 32;
	    intmode = BIN;
	    msg("Binary mode");
	}
	break;
	
    case OCT:
	if (mode != PROG || intmode != OCT) {
	    stop_entering();
	    intmode = OCT;
	    mode = PROG;
	    os_raw_mode(0);
	    max_digits = 11;
	    msg("Octal mode");
	}
	break;
	
    case DEC:
	if (mode != PROG || intmode != DEC) {
	    mode = PROG;
	    stop_entering();
	    intmode = DEC;
	    os_raw_mode(0);
	    max_digits = 10;
	    msg("Decimal mode");
	}
	break;
	
    case HEX:
	if (mode != PROG || intmode != HEX) {
	    mode = PROG;
	    stop_entering();
	    intmode = HEX;
	    os_raw_mode(0);
	    max_digits = 8;
	    msg("Hexadecimal mode");
	}
	break;
	
    case FIN:
    case SCI:
    case STAT:
	if (mode != c) {
	    char buf[80];
	    mode = c;
	    stop_entering();
	    os_raw_mode(0);
	    max_digits = 20;
	    sprintf(buf, "%s mode", (c==FIN)? "Financial": (c==SCI)? "Scientific": "Statistics");
	    msg(buf);
	}
	break;

    case PROG:
	if (mode != c) {
	    mode = PROG;
	    stop_entering();
	    os_raw_mode(0);
	    max_digits = 10;
	    msg("Programmers mode");
	}
	break;

    default:
	put_a_char(BELL);
	break;
    }
    dispnums();
    dispregs();
    prinbase();
}

static void addPrime(long cand, long *primes, int *primeMultiplier, int *numPrimes) {
    if (*numPrimes == 0 ) {
	*numPrimes = 1;
	primes[0] = cand;
	primeMultiplier[0] = 1;
    } else {
	if (primes[*numPrimes - 1] == cand) {
	    primeMultiplier[*numPrimes - 1] = primeMultiplier[*numPrimes - 1] + 1;
	} else {
	    if (*numPrimes == KMaxPrimes) {
		msg("Too many primes");
		return;
	    }
	    primes[*numPrimes] = cand;
	    primeMultiplier[*numPrimes] = 1;
	    (*numPrimes)++;
	}
    }
}

static void factors(long orig) {
    long 	target;
    int		i;
    char	msgbuf[100];
    long	cand, rem, term;
    long 	primes[KMaxPrimes];
    int 	primeMultiplier[KMaxPrimes], numPrimes;

    if (orig < 2) {
	msg("No primes");
	return;
    }

    target = orig;
    term = sqrt((double) target) + 1;
    numPrimes = 0;
    for (i = 0; i < KMaxPrimes; i++) {
	primes[i] = 0;
	primeMultiplier[i] = 0;
    }

    // check for 2's
    cand = 2;
    while (cand < term) {
	rem = target / cand;
	if (rem * cand == target) {
	    target = rem;
	    term = sqrt(target) + 1;
	    addPrime(cand, primes, primeMultiplier, &numPrimes);
	} else
	    break;
    }

    // check odd numbers
    cand = 3;
    while (cand < term) {
	rem = target / cand;
	if (rem * cand == target) {
	    target = rem;
	    term = sqrt(target)+1;
	    addPrime(cand, primes, primeMultiplier, &numPrimes);
	} else
	    cand += 2;
    }
    addPrime(target, primes, primeMultiplier, &numPrimes);

    if (numPrimes == 1) {
	sprintf(msgbuf, "%ld is prime", orig);
    } else {
	sprintf(msgbuf,"%ld = ", orig);
	for (i=0, target = 1; i < numPrimes; i++) {
	    char *m = msgbuf;
	    int j;

	    m += strlen(msgbuf);
	    for (j = 0; j < primeMultiplier[i]; j++)
		target *= primes[i];
	    sprintf(m, " %ld", primes[i]);
	    if (primeMultiplier[i] > 1) {
		m += strlen(m);
		sprintf(m, "**%d", primeMultiplier[i]);
	    }
	    if (i < numPrimes - 1) {
		strcat(m, " x ");
	    }
	}
	msg(msgbuf);
    }

    if (target != orig) { // ERRORCHECKING IN BETA ONLY!!!
	msg("Error!");
    }
}

static int power(int x, int n) {
    int i, p;
    p = 1;
    for (i = 1; i <= n; ++i)
	p *= x;
    return p;
}

static void arith_int(COMMAND c) {
    long	tempx, tempy;
    char 	*c_ptr, c_buf[45];
    
    if (c == CHS) { /* CHS */
	if (entering) {
	    c_ptr = inbuf;
	    strcpy(c_buf, c_ptr);
	    *c_ptr = ASCIIZ;
	    c_ptr = c_buf;
	    if (*c_ptr == '-')
		c_ptr++;
	    else
		strcat(inbuf, "-");
	    strcat(inbuf, c_ptr);
	    inptr = &inbuf[strlen(inbuf)];
	} else {
	    tempx = pop(); 
	    push(-tempx); 
	}
	dispnums();
	return;
    }
    
    if (c == '@') { /* Pop up registers window */
	pop_up_reg();
	return;
    }
    
    save_x();
    tempx = pop();
    
    switch (c) {
	/* operations on X only */
    case NOT: 
	tempx = ~tempx;
	break;

    case SHIFTL:
	tempx <<= 1;
	break;

    case SHIFTR:
	tempx >>= 1;
	break;

    case PRIMEF:
	factors(tempx);
	break;
	
    case SQR:
	tempx *= tempx;
	break;

    default: { /* X & Y operations */
	tempy = pop(); 
	switch (c) {
	case '+':
	case PLUS:
	    tempx = tempx + tempy;
	    break;
	case '-':
	case MINUS:
	    tempx = tempy - tempx;
	    break;
	case '*':
	case MULT:
	    tempx = tempx * tempy;
	    break;
	case '/':
	case DIVIDE:
	    if (tempx == 0) {
		push(tempy);
		msg(div_by_zero);
	    } else {
		if (tempx != 0)
		    tempx = tempy / tempx;
		else
		    tempx = 0;
	    }
	    break;

	case '^':
	case XOR:
	    tempx ^= tempy;
	    break;
	case '|':
	case OR:
	    tempx |= tempy;
	    break;
	case '&':
	case AND:
	    tempx &= tempy;
	    break;
	case '<':
	case SHIFTYL:
	    tempy <<= tempx;
	    tempx = tempy;
	    break;
	case '>':
	case SHIFTYR:
	    tempy >>= tempx;
	    tempx = tempy;
	    break;
	case MODULUS:
	    tempy %= tempx;
	    tempx = tempy;
	    break;
	case YTOX:
	    if (tempx < 0) {
		msg("X cannot be < 0");
		push(tempy);
	    } else
		tempx = power(tempy, tempx);
	    break;

	default:
	    push(tempy);
	    put_a_char(BELL);
	    break;
	}
    }
	push(tempx);
	dispnums();
    }
}

static void arith_real(COMMAND c) {
    double	tempxf, tempyf;
    char 	*c_ptr, c_buf[45];
    
    if (c == CHS) { /* CHS */
	if (entering) {
	    c_ptr = inbuf;
	    if (strchr(inbuf, '.') || strchr(inbuf, 'E')) {
		while ((*c_ptr != 'E') && (*c_ptr != 'e'))
		    c_ptr++;
		c_ptr++;
	    }
	    strcpy(c_buf, c_ptr);
	    *c_ptr = ASCIIZ;
	    c_ptr = c_buf;
	    if (*c_ptr == '-')
		c_ptr++;
	    else
		strcat(inbuf, "-");
	    strcat(inbuf, c_ptr);
	    inptr = &inbuf[strlen(inbuf)];
	} else {
	    tempxf = popf();
	    pushf(-tempxf);
	}
	dispnums();
	return;
    }
    
    if (c == '@') { /* Pop up registers window */
	pop_up_reg();
	return;
    }
    
    save_x();
    tempxf = popf();
    
    switch (c) {
	/* operations on X only */

    case RECI:
	if (tempxf == 0.0)
	    msg(div_by_zero);
	else {
	    tempxf = 1.0 / tempxf;
	}
	break;

    default: { /* X & Y operations */
	tempyf = popf();
	switch (c) {
	case '+':
	case PLUS:
	    tempxf = tempxf + tempyf;
	    break;
	case '-':
	case MINUS:
	    tempxf = tempyf - tempxf;
	    break;
	case '*':
	case MULT:
	    tempxf = tempxf * tempyf;
	    break;
	case '/':
	case DIVIDE:
	    if (tempxf == 0.0) {
		pushf(tempyf);
		msg(div_by_zero);
	    } else {
		tempxf = tempyf / tempxf;
	    }
	    break;

	case '%':
	case PERCENT:
	    pushf(tempyf); /* push y back on stack for future ops */
	    if (isInverted()) {
		if (tempxf == 0.0) {
		    msg(div_by_zero);
		    pushf(tempxf);
		} else {
		    tempxf = tempyf * 100.0 / tempxf;
		}
	    } else {
		tempxf = tempxf * tempyf / 100.0;
	    }
	    break;

	case PERCENTCH:
	    pushf(tempyf); /* push y back on stack for future ops */
	    if (isInverted()) {
		if (tempxf == 0.0) {
		    msg(div_by_zero);
		    pushf(tempxf);
		} else {
		    tempxf = tempyf * (1.0 + tempxf / 100.0);
		}
	    } else {
		tempxf = 100.0 * (tempxf - tempyf) / tempyf;
	    }
	    break;

	case YTOX:
	    tempxf = pow(tempyf, tempxf);
	    break;

	default:
	    pushf(tempyf);
	    put_a_char(BELL);
	    break;
	}
    }
    }
    pushf(tempxf);
    dispnums();
}

static void trig(COMMAND c) {
    double	tempxf, tempyf;
    
    /* TRIG FUNCTIONS ... */
    save_x();
    tempxf = popf();
    
    switch(c) {
    case SIN:
	if (isInverted()) {
	    tempxf = asin(tempxf);
	    if (degree)
		tempxf *= r_to_d;
	} else {
	    if (degree)
		tempxf = sin(d_to_r * tempxf);
	    else
		tempxf = sin(tempxf);
	}
	break;
	
    case COS:
	if (isInverted()) {
	    tempxf = acos(tempxf);
	    if (degree)
		tempxf *= r_to_d;
	} else {
	    if (degree)
		tempxf = cos(d_to_r * tempxf);
	    else
		tempxf = cos(tempxf);
	}
	break;
	
    case TAN:
	if (isInverted()) {
	    tempxf = atan(tempxf);
	    if (degree)
		tempxf *= r_to_d;
	} else {
	    if (degree)
		tempxf = tan(d_to_r * tempxf);
	    else
		tempxf = tan(tempxf);
	}
	break;

    case SINH:
	if (isInverted()) {
	    tempxf = log(tempxf+sqrt(tempxf*tempxf + 1));
	} else {
	    tempxf = 0.5*(exp(tempxf) - exp(-tempxf));
	}
	break;

    case COSH:
	if (isInverted()) {
	    if (tempxf < 1) {
		msg("Illegal value");
		pushf(tempxf);
	    } else {
		tempxf = log(tempxf+sqrt(tempxf*tempxf - 1));
	    }
	} else {
	    tempxf = 0.5*(exp(tempxf) + exp(-tempxf));
	}
	break;

    case TANH:
	if (isInverted()) {
	    if (tempxf == 1) {
		msg("Illegal value");
		pushf(tempxf);
	    } else {
		tempxf = 0.5*log((tempxf+1)/(1-tempxf));
	    }
	} else {
	    double p;
	    p = exp(2*tempxf);
	    tempxf = (p-1)/(p+1);
	}
	break;

    case ETOX:
	if (isInverted()) {
	    tempxf = log(tempxf);
	} else {
	    tempxf = exp(tempxf);
	}
	break;
	
    case LOGE:
	if (isInverted()) {
	    tempxf = exp(tempxf);
	} else {
	    tempxf = log(tempxf);
	}
	break;
	
    case LOG10:
	if (isInverted()) {
	    tempxf = pow(10.0, tempxf);
	} else {
	    tempxf = log10(tempxf);
	}
	break;
	
	/* Square root keys ... */
	
    case SQR:
	tempxf *= tempxf;
	break;
	
    case ROOT:
	tempxf = sqrt(tempxf);
	break;
	
    case CUBE:
	tempxf = tempxf * tempxf * tempxf;
	break;
	
    case CROOT:
	/* Floating point part ... */
	tempxf = pow(tempxf, 1.0 / 3.0);
	break;
	
    case FRC:
    {
	int tempy = 0;
	if (tempxf < 0.0) {
	    tempy = 1;
	    tempxf = -tempxf;
	}
	tempxf = tempxf - floor(tempxf);
	if (tempy)
	    tempxf = -tempxf;
	break;
    }
	
    case INT:
    {
	int tempy = 0;
	if (tempxf < 0.0) {
	    tempy = 1;
	    tempxf = -tempxf;
	}
	tempxf = (double) floor(tempxf);
	if (tempy)
	    tempxf = -tempxf;
	break;
    }

    case HMS:
	if (isInverted()) { /* convert h.mmss to hours */
	    long h, m, s;
	    int sign = 0;

	    if (tempxf < 0.0) {
		sign = 1;
		tempxf = -tempxf;
	    }

	    /* round up a 1/100000000th of an hour for corner cases (e.g. 43d30m00s) */
	    tempxf += 1.0E-8;

	    h = tempxf;
	    tempxf = (tempxf - h) * 100;
	    m = tempxf;
	    tempxf = (tempxf - m) * 100;
	    s = tempxf;
	    tempxf = h + m / 60.0 + s / 3600.0;
	    if (sign)
		tempxf = -tempxf;
	    msg("HH,MMSS -> hours");
	} else { /* convert hours to h.mmss */
	    long h, m, s;
	    int sign = 0;

	    if (tempxf < 0.0) {
		sign = 1;
		tempxf = -tempxf;
	    }
	    h = tempxf;
	    tempxf = (tempxf - h) * 60;
	    m = tempxf;
	    s = (tempxf - m) * 60;
	    tempxf = h + m / 100.0 + s / 10000.0;
	    if (decplaces < 4)
		decplaces = 4;
	    if (sign)
		tempxf = -tempxf;
	    msg("hours -> HH,MMSS");
	}
	break;

    case DTOR:
	if (isInverted()) { 
	    tempxf *= d_to_r;
	    msg("degrees to radians");
	} else {
	    tempxf *= r_to_d;
	    msg("radians to degrees");
	}
	break;
	
    case RTOP:
	tempyf = popf();
	if (isInverted()) { /*  convert (r, theta) to (x, y) */
	    double theta, r;

	    theta = tempyf;
	    if (degree)
		theta = d_to_r * theta;

	    r = tempxf;
	    tempyf = r * sin(theta);
	    tempxf = r * cos(theta);
	    msg("polar to rectangular");
	} else { /*  convert (x, y) to (r, theta) */
	    double theta, r;

	    r = sqrt(tempxf * tempxf + tempyf * tempyf);
	    theta = atan2(tempyf, tempxf);
	    if (degree)
		theta = r_to_d * theta;
	    tempyf = theta;
	    tempxf = r;
	    msg("rectangular to polar");
	}
	pushf(tempyf);
	break;

    default:
	put_a_char(BELL);
	break;
    }
    pushf(tempxf);
    dispnums();
}

static double fact(double from, double to) {
    double total, n, m;

    n = floor(from);
    m = floor(to);
    if (n <= 0 || m <= 0)
	return(1.0);

    total = 1.0;
    while (n > m) {
	total = total * n;
	n = n - 1;
    }
    return total;
}

static void summation(COMMAND c) {
    double	tempxf, tempyf;
    
    switch(c) {
    case FACT:
	save_x();
	tempxf = popf();
	if (tempxf < 0) {
	    msg("Illegal value");
	    pushf(tempxf);
	} else {
	    pushf(fact(tempxf, 1));
	    dispnums();
	    
	}
	break;

    case SUM:
	save_x();
	tempxf = popf();
	tempyf = popf();
	pushf(tempyf);
	pushf(tempxf);
	regf[STAT_NUM_REG]++;
	regf[STAT_SUMX_REG] += tempxf;
	regf[STAT_SUMX2_REG] += tempxf * tempxf;
	regf[STAT_SUMY_REG] += tempyf;
	regf[STAT_SUMY2_REG] += tempyf * tempyf;
	regf[STAT_SUMXY_REG] += tempxf * tempyf;
	pushf(regf[STAT_NUM_REG]);
	dispreg(STAT_NUM_REG);
	dispreg(STAT_SUMX_REG);
	dispreg(STAT_SUMY_REG);
	dispreg(STAT_SUMX2_REG);
	dispreg(STAT_SUMY2_REG);
	dispreg(STAT_SUMXY_REG);
	dispnums();
	msg("X & Y accumulated");
	break;
	
    case SUMR:
	if (regf[STAT_NUM_REG] > 0) {
	    save_x();
	    tempxf = popf();
	    tempyf = popf();
	    pushf(tempyf);
	    pushf(tempxf);
	    regf[STAT_NUM_REG]--;
	    regf[STAT_SUMX_REG] -= tempxf;
	    regf[STAT_SUMX2_REG] -= tempxf * tempxf;
	    regf[STAT_SUMY_REG] -= tempyf;
	    regf[STAT_SUMY2_REG] -= tempyf * tempyf;
	    regf[STAT_SUMXY_REG] -= tempxf * tempyf;
	    pushf(regf[STAT_NUM_REG]);
	    dispreg(STAT_NUM_REG);
	    dispreg(STAT_SUMX_REG);
	    dispreg(STAT_SUMY_REG);
	    dispreg(STAT_SUMX2_REG);
	    dispreg(STAT_SUMY2_REG);
	    dispreg(STAT_SUMXY_REG);
	    dispnums();
	    msg("X & Y decremented");
	}
	break;
	
    case SUM0:
	regf[STAT_NUM_REG] = 0.0;
	regf[STAT_SUMX_REG]  = 0.0;
	regf[STAT_SUMX2_REG]  = 0.0;
	regf[STAT_SUMY_REG]  = 0.0;
	regf[STAT_SUMY2_REG]  = 0.0;
	regf[STAT_SUMXY_REG]  = 0.0;
	dispreg(STAT_NUM_REG);
	dispreg(STAT_SUMX_REG);
	dispreg(STAT_SUMY_REG);
	dispreg(STAT_SUMX2_REG);
	dispreg(STAT_SUMY2_REG);
	dispreg(STAT_SUMXY_REG);
	dispnums();
	msg("Accumulator registers cleared");
	break;
	
    case NSTAT:
	pushf(regf[STAT_NUM_REG]);
	dispnums();
	break;
	
    case MEAN:
	if (regf[STAT_NUM_REG] == 0.0)
	    msg(div_by_zero);
	else {
	    tempxf = regf[STAT_SUMX_REG] / regf[STAT_NUM_REG];
	    pushf(tempxf);
	    dispnums();
	    msg("Mean of X's");
	}
	break;
	
    case MEANY:
	if (regf[STAT_NUM_REG] == 0.0)
	    msg(div_by_zero);
	else {
	    tempyf = regf[STAT_SUMY_REG] / regf[STAT_NUM_REG];
	    pushf(tempyf);
	    dispnums();
	    msg("Mean of Y's");
	}
	break;
	
    case S_DEV:
	/*  s.dev = sqrt((sum:(x^2) - (sum:(x))^2 /n) / n-1) */
	if (regf[STAT_NUM_REG] == 0.0)
	    msg(div_by_zero);
	else {
	    tempyf = sqrt((regf[STAT_SUMY2_REG] - regf[STAT_SUMY_REG]*regf[STAT_SUMY_REG]/regf[STAT_NUM_REG])/(regf[STAT_NUM_REG] - 1));
	    pushf(tempyf);
	    tempxf = sqrt((regf[STAT_SUMX2_REG] - regf[STAT_SUMX_REG]*regf[STAT_SUMX_REG]/regf[STAT_NUM_REG])/(regf[STAT_NUM_REG] - 1));
	    pushf(tempxf);
	    dispnums();
	    msg("Std Dev of X's & Y's");
	}
	break;

    case PERM:
	save_x();
	tempxf = popf();
	tempyf = popf();
	if (tempxf < 0 || tempyf < 0 || tempyf < tempxf) {
	    msg("Bad values");
	    pushf(tempyf);
	    pushf(tempxf);
	} else {
	    char buf[80];
	    double n, r;

	    n = floor(tempyf+.5);
	    r = floor(tempxf+.5);
	    pushf(fact(n, n - r));	
	    dispnums();
	    sprintf(buf, "Permutations (%.0f, %.0f)", n, r);
	    msg(buf);
	}
	break;

    case COMB:
	save_x();
	tempxf = popf();
	tempyf = popf();
	if (tempxf < 0 || tempyf < 0 || tempyf < tempxf) {
	    msg("Bad values");
	    pushf(tempyf);
	    pushf(tempxf);
	} else {
	    double n, r;
	    char buf[80];

	    n = floor(tempyf+.5);
	    r = floor(tempxf+.5);
	    if (n-r > r)
		pushf(tempxf = fact(n, n-r) / fact(r, 1));
	    else
		pushf(tempxf = fact(n, r) / fact(n-r, 1));
	    dispnums();
	    sprintf(buf, "Combinations (%.0f, %.0f)", n, r);
	    msg(buf);
	}
	break;

    case LR:
	if (isInverted()) {
	    char buf[80];

	    sprintf(buf, "Linear Regression formula is: Y= %f*X + %f",
		    regf[STAT_SLOPE], regf[STAT_Y_INT]);
	    msg(buf);
	} else {
	    if (regf[STAT_NUM_REG] < 2)
		msg("Not enough data points");
	    else {
		double detx, dety, a, b, r2;

		detx = regf[STAT_SUMX2_REG]*regf[STAT_NUM_REG] - 
		    regf[STAT_SUMX_REG]*regf[STAT_SUMX_REG];
		dety = regf[STAT_SUMY2_REG]*regf[STAT_NUM_REG] - 
		    regf[STAT_SUMY_REG]*regf[STAT_SUMY_REG];
		a = (regf[STAT_SUMXY_REG]*regf[STAT_NUM_REG] - 
		     regf[STAT_SUMX_REG]*regf[STAT_SUMY_REG])/detx;
		b = (regf[STAT_SUMX2_REG]*regf[STAT_SUMY_REG] - 
		     regf[STAT_SUMX_REG]*regf[STAT_SUMXY_REG])/detx;
		r2 = a*a*detx/dety;
		regf[STAT_SLOPE] = a;
		regf[STAT_Y_INT] = b;
		regf[STAT_R2] = b;
		pushf(r2);
		pushf(b);
		pushf(a);
		dispnums();
		msg("Slope (X), Y-intersect (Y) & correlation (Z) calculated");
	    }
	}
	break;
		
    case CALCY:
	if (isInverted()) {
	    tempyf = popf();
	    tempxf = (tempyf -  regf[STAT_Y_INT]) / regf[STAT_SLOPE];
	    msg("X calculated from Y");
	} else {
	    tempxf = popf();
	    tempyf = regf[STAT_SLOPE] * tempxf + regf[STAT_Y_INT];
	    msg("Y calculated from X");
	} 
	pushf(tempyf);
	pushf(tempxf);
	dispnums();
	break;

    default:
	put_a_char(BELL);
	break;
    }
}

static int dayFunction(double f) {
    long y, m, d, x, z;
	
    y = f;
    f = (f - y) * 100;
    m = f;
    f = (f - m) * 100;
    d = f;
    if (m <= 2) {
	x = 0;
	z = y - 1;
    } else {
	x = (long) (0.4*m + 2.3);
	z = y;
    }
    return 365*y + 31 *(m - 1) + d + (int)(z/4) - x;
}

static int dayFunction30(double f, int i) {
    static int was30or31;
    int y, m, d, a1;
	
    y = f;
    f = (f - y) * 100;
    m = f;
    f = (f - m) * 100;
    d = f;
	
    if (i == 1) {
	if (d == 31) {
	    a1 = 30;
	} else {
	    a1 = d;
	}
	if (d == 30 || d == 31) {
	    was30or31 = 1;
	} else {
	    was30or31 = 0;
	}
    } else {
	if (d == 31) {
	    if (was30or31) {
		a1 = 30;
	    } else {
		a1 = d;
	    }
	} else {
	    a1 = d;
	}
    }
    return 360*y + 30*m + a1;
}

static void financial(COMMAND c) {
    /*
     * Financial function register allocation ...
     *
     * 5 n
     * 6 interest
     * 7 present value
     * 8 payment
     * 9 future value
     *
     * Formulae are :
     *
     * 0 = PVAL + (1 + I.S) PMT (1 - (1+I)**-N ) / I + FVAL * (1+I)**-N

     New Formula from psion version:
     y=Fval + Pval*(1 + i)**n + (1 + s)*pmt*((1 + i)**n - 1)/i
     y'=pmt*(1 + (1 + i)^(n-1) * n * (pval* i^2 / pmt - (1+i)/n + i + s*i^2)) / i^2
     *
     * where:
     *
     * N    = number of periods
     * PVAL = present value
     * I    = interest rate (as a fraction - user sees a percent)
     * PMT  = payment
     * FVAL = future value
     * S    = payment mode factor. 0 for payment at end of period, 1 for start.
     *        This version only uses S=0.
     */
    
    double	tempxf, tempyf;
    int 	count;
    BOOLEAN	this_was_fin_key;
    double 	interest, n, pval, pmt, fval, guess, epsilon, dy, fy;
    
    this_was_fin_key = FALSE;
    switch (c) {
    case PVAL:
	if (isInverted()) {
	    pushf(regf[FIN_PVAL_REG]);
	    dispnums();
	    msg("Present value recalled");
	} else if (!last_was_fin_key) {
	    save_x();
	    tempxf = popf();
	    pushf(tempxf);
	    regf[FIN_PVAL_REG] = tempxf;
	    dispreg(FIN_PVAL_REG);
	    dispnums();
	    this_was_fin_key = TRUE;
	    msg("Present value entered");
	} else {
	    n = regf[FIN_NUM_REG];
	    interest = regf[FIN_INT_REG] / 100.0;
	    pmt = regf[FIN_PMT_REG];
	    fval = regf[FIN_FVAL_REG];
	    if (interest == 0.0) {
		tempxf = -(fval + (n + finPayAt0)*pmt);
	    } else {
		tempyf = pow(1.0 + interest, -n);
		tempxf = -(fval*tempyf + (1 + finPayAt0*interest)*pmt*(1.0 - tempyf)/interest);
	    }
	    pushf(tempxf);
	    regf[FIN_PVAL_REG] = tempxf;
	    dispreg(FIN_PVAL_REG);
	    dispnums();
	    msg("Present value computed");
	}
	break;
	
    case FVAL:
	if (isInverted()) {
	    pushf(regf[FIN_FVAL_REG]);
	    dispnums();
	    msg("Future value recalled");
	} else if (!last_was_fin_key) {
	    save_x();
	    tempxf = popf();
	    pushf(tempxf);
	    regf[FIN_FVAL_REG] = tempxf;
	    dispreg(FIN_FVAL_REG);
	    dispnums();
	    this_was_fin_key = TRUE;
	    msg("Final value entered");
	} else {
	    n = regf[FIN_NUM_REG];
	    interest = regf[FIN_INT_REG] / 100.0;
	    pval = regf[FIN_PVAL_REG];
	    pmt = regf[FIN_PMT_REG];
	    if (interest == 0.0) {
		tempxf = -(pval + (n + finPayAt0) * pmt);
	    } else {
		tempyf = pow(1.0 + interest, n);
		tempxf = -(pval*tempyf + (1 + finPayAt0*interest)*pmt*(tempyf - 1.0)/interest);
	    }
	    pushf(tempxf);
	    regf[FIN_FVAL_REG] = tempxf;
	    dispreg(FIN_FVAL_REG);
	    dispnums();
	    msg("Final value calculated");
	}
	break;
	
    case PMT:
	if (isInverted()) {
	    pushf(regf[FIN_PMT_REG]);
	    dispnums();
	    msg("Payment recalled");
	} else if (!last_was_fin_key) {
	    save_x();
	    tempxf = popf();
	    pushf(tempxf);
	    regf[FIN_PMT_REG] = tempxf;
	    dispreg(FIN_PMT_REG);
	    dispnums();
	    this_was_fin_key = TRUE;
	    msg("Payment entered");
	} else {
	    n = regf[FIN_NUM_REG];
	    interest = regf[FIN_INT_REG] / 100.0;
	    pval = regf[FIN_PVAL_REG];
	    fval = regf[FIN_FVAL_REG];
	    if (fabs(interest) < 1.0e-13) {
		tempxf = -(fval + pval)/(n + finPayAt0);
	    } else {
		tempyf = pow(1.0 + interest, n);
		tempxf = -interest*(fval + pval*tempyf)/((1.0 + interest*finPayAt0)*(tempyf - 1.0));
	    }
	    pushf(tempxf);
	    regf[FIN_PMT_REG] = tempxf;
	    dispreg(FIN_PMT_REG);
	    dispnums();
	    msg("Payment calculated");
	}
	break;
	
    case INTST:
	if (isInverted()) {
	    pushf(regf[FIN_INT_REG]);
	    dispnums();
	    msg("Interest recalled");
	} else if (!last_was_fin_key) {
	    save_x();
	    tempxf = popf();
	    pushf(tempxf);
	    regf[FIN_INT_REG] = tempxf;
	    dispreg(FIN_INT_REG);
	    dispnums();
	    this_was_fin_key = TRUE;
	    msg("Interest rate (%) entered");
	} else {
	    n = regf[FIN_NUM_REG];
	    pval = regf[FIN_PVAL_REG];
	    pmt = regf[FIN_PMT_REG];
	    fval = regf[FIN_FVAL_REG];
	    if (fabs(fval + pval + (n + finPayAt0)*pmt) == 0.0) {
		tempxf = 0.0;
		epsilon = 0.0;
	    } else {
		/* use Newton-Raphson iteration ...
		   if we are seeking roots of f(x)=0 then
		   x_n+1 = x_n - f(x_n) / f'(x_n)
		   y=-fval + pval*(1+i)**n + pmt*((1+i)**n - 1))/i
		   y'=(pval + pmt/i)*n*(1+i)**(n-1) + pmt*(1-(1+i)**n)/i**2
		*/
		count = 0;
		guess = 0.01; /* == 12% per annum */
		if (fabs(fval - pval - n * pmt) < 1.0e-12) {
		    tempxf = 0.0;
		    epsilon = 0.0;
		} else {
		    epsilon = 1000.0;
		    while ((count < 10000) && (fabs(epsilon) > 1.0e-13)) {
			tempyf = pow(1.0 + guess, n);
			fy = fval + pval*tempyf + (1.0 + guess*finPayAt0)*pmt*(tempyf - 1.0)/guess;
			dy = (((pval + pmt*finPayAt0)*n*guess*guess + (n - 1)*guess*pmt - pmt)*pow(1+guess,n-1) + pmt)/(guess*guess);
			epsilon=fy/dy;
			guess -= epsilon;
			count++;
		    }
		}
		tempxf = guess * 100.0; /* Convert to percent */
		if (count < 10000) {
		    pushf(tempxf);
		    regf[FIN_INT_REG] = tempxf;
		    dispreg(FIN_INT_REG);
		    dispnums();
		    msg("Interest rate (%) computed");
		} else
		    msg("Calculation does not converge!\007");
	    }
	}
	break;
	
    case NFIN:
	if (isInverted()) {
	    pushf(regf[FIN_NUM_REG]);
	    dispnums();
	    msg("Number of payments recalled");
	} else if (!last_was_fin_key) {
	    save_x();
	    tempxf = popf();
	    pushf(tempxf);
	    regf[FIN_NUM_REG] = tempxf;
	    dispreg(FIN_NUM_REG);
	    dispnums();
	    this_was_fin_key = TRUE;
	    msg("Number of payments entered");
	} else {
	    interest = regf[FIN_INT_REG] / 100.0;
	    pval = regf[FIN_PVAL_REG];
	    pmt = regf[FIN_PMT_REG];
	    fval = regf[FIN_FVAL_REG];
	    if (interest == 0.0)
		tempxf = -(fval + pval)/pmt - finPayAt0;
	    else {
		tempyf = 1.0 + interest*finPayAt0;
		tempxf = log((tempyf*pmt/interest - fval)/(pval + tempyf*pmt/interest))/log(1.0 + interest);
	    }
	    pushf(tempxf);
	    regf[FIN_NUM_REG] = tempxf;
	    dispreg(FIN_NUM_REG);
	    dispnums();
	    msg("Number of payments computed");
	}
	break;

    case FCLR:
	c = dialog("Clear all finance registers? (Y or N)");
	if (toupper(c) == 'Y') {
	    int i;
	    for (i=5; i< 10; i++)
		regf[i] = 0;
	}
	break;

    case BEGIN:
	if (finPayAt0) {
	    finPayAt0 = 0;
	    msg("Annuity in advance");
	} else {
	    finPayAt0 = 1;
	    msg("Annuity in arrears");
	}
	break;

    case DYS:
	save_x();
	tempxf = popf();
	tempyf = popf();
	if (isInverted()) {
	    tempyf = dayFunction30(tempxf,1);
	    tempxf = dayFunction30(tempyf,2) - tempyf;
	} else {
	    tempxf = dayFunction(tempyf) - dayFunction(tempxf);
	}
	msg("days between two dates calculated");
	if (decplaces < 4)
	    decplaces = 4;
	pushf(tempxf);
	dispnums();
	break;

    case TDY:
    { 
	struct tm *tmbuf;
	time_t t;

	t = time(NULL);
	tmbuf = localtime(&t);
	tempxf = tmbuf->tm_year + 1900;
	tempxf += (tmbuf->tm_mon + 1) / 100.0;
	tempxf += tmbuf->tm_mday / 10000.0;
	pushf(tempxf);
	if (decplaces < 4)
	    decplaces = 4;
	dispnums();
	msg("Today's date");
	break;
    }

    case TIMES12:
	save_x();
	tempxf = popf();
	tempxf *= 12.0;
	pushf(tempxf);
	dispnums();
	break;

    case DIVIDE12:
	save_x();
	tempxf = popf();
	tempxf /= 12.0;
	pushf(tempxf);
	dispnums();
	break;

    default:
	put_a_char(BELL);
	break;
    }
    last_was_fin_key = FALSE;
    if (this_was_fin_key)
	last_was_fin_key = TRUE;
}

static void process_digit(COMMAND c) {
    switch (c) {
    case BACKSPACE:
    case '0':
    case '1':
	echo_char(c);
	break;
	
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
	if (mode == PROG && intmode == BIN)
	    put_a_char(BELL);
	else
	    echo_char(c);
	break;
	
    case '8':
    case '9':
	if ((mode == PROG) && ((intmode == BIN) || (intmode == OCT)))
	    put_a_char(BELL);
	else
	    echo_char(c);
	break;
	
    case 'A':
    case 'B':
    case 'C':
    case 'D':
    case 'F':
	if (mode != PROG || intmode != HEX)
	    put_a_char(BELL);
	else
	    echo_char(c);
	break;
	
    case 'E':
	if (mode == PROG) {
	    if (intmode == HEX) {
		echo_char(c);
	    } else {
		put_a_char(BELL);
	    }
	} else {
	    if (strchr(inbuf, 'E') == NULL) {
		if (strlen(inbuf) == 0)
		    echo_char('1');
		echo_char(c);
	    } else {
		msg("Can't put another E in the number!");
		put_a_char(BELL);
	    }
	}
	break;
	
    case '.':
	if (mode == PROG) { /* ASCIIM never reaches here! */
	    if (intmode == IP) {
		echo_char('.');
	    } else {
		put_a_char(BELL);
	    }
	} else if (entering && strchr(inbuf, '.'))
	    put_a_char(BELL);
	else
	    echo_char(c);
	break;
	
    default:
	put_a_char(BELL);
	break;
    }
}

/* trim all white space from buf (for units) */
static void trim(char *s) {
    char *d = s;
    while (*s) {
	if (!isspace(*s))
	    *d++ = *s;
	s++;
    }
    *d = 0;
}

void convertX(char *from, char *to) {
    int fd[2], pid;

    if (mode == PROG) {
	msg("No conversions in PROG mode");
	return;
    }

    trim(from);
    trim(to);

    stop_entering();
    if (pipe(fd) < 0)
	msg("Can't open a pipe.");
    else {
	double oldX = xf;
	char *argv[5];
	char fromarg[80];

	if (fabs(oldX) < 0.0001)
	    oldX = 1.0;

	sprintf(fromarg, "%f %s", oldX, from);
	argv[0] = "units";
	argv[1] = "-o %.20g";
	argv[2] = fromarg;
	argv[3] = to;
	argv[4] = 0;
	if ((pid = vfork()) == -1) {
	    msg("Fork failed.");
	} else if (pid == 0) { /* child - put stdout on pipe */
	    close(fd[0]); /* child does not need to read from pipe */
	    close(1); /* set stdout to the pipe */
	    dup(fd[1]);
	    close(fd[1]);
	    execvp(argv[0], argv);
	    _exit(1); /* abandon hope all ye who get here */
	} else { /* PARENT - read from pipe */
	    FILE *p;
	    close(fd[1]); /* parent does not write to pipe */
	    if ((p = fdopen(fd[0], "r")) == NULL) {
		msg("Can't read from pipe");
	    } else {
		double newX;
		int numRead = 0, status, bytesRead = 0;
		char inbuf[80];

		while (fgets(inbuf, 80, p) != NULL) {
		    bytesRead += strlen(inbuf);
		    if (!numRead)
			numRead = sscanf(inbuf, " * %lf", &newX);
		}
		wait(&status);
		if (numRead == 1 && 
		    WIFEXITED(status) && 
		    WEXITSTATUS(status) == 0) {
		    double xf;

		    xf = popf();
		    pushf(newX);
		    if (fabs(xf) < 0.0001)
			sprintf(inbuf, "Conversion factor %s to %s", from, to);
		    else
			sprintf(inbuf, "Converted %s to %s", from, to);
		    msg(inbuf);
		    dispnums();
		} else if (bytesRead)
		    /* Note: if you are tempted to catch units's error
		       message be aware that as of version 1.55 it
		       does not use stderr! */
		    msg("Bad units");
		else
		    msg("Can't find the units program - please install it!");

		fclose(p); /* no need to close fd[1] too */
	    }
	}
    }
}

void iProcess(COMMAND c, int i) {
    switch (c) {
    case RECALL:
	if ((i >= 0) && (i < NUMREGS)) {
	    char buf[80];

	    if (mode == PROG) {
		push(reg[i]); 
	    } else {
		pushf(regf[i]);
	    }
	    dispnums();
	    sprintf(buf, "Recalled from R%d", i);
	    msg(buf);
	} else
	    msg("index out of bounds");
	break;
    default:
	break;
    }
}

int process(COMMAND c) {
    double	tempxf, tempyf;
    long	tempx, tempy;
    int	i, j;
    extern 	int is_resident;
    
    clear_msg();

    /* ASCII MODE - almost any character can be echo'ed */
    if ((c < MIN_COMMAND) && (mode == PROG && intmode == ASCIIM)) {
	echo_char(c); /* An ordinary character was typed */
	return 0;
    }
    
    if (c == INV) {
	invert = !invert;
	print_inv();
	clear_msg();
	return 0;
    }    
    
    if (((c >= '0') && (c <= '9')) ||
	((c >= 'A') && (c <= 'F')) ||
	(c == BACKSPACE) ||
	(c == '.')) {
	process_digit(c);
	return 0;
    }
    
    if (((c >= '!') && (c <= '/')) ||
	((c >= ':') && (c <= '@')) ||
	((c >= '[') && (c <= '`')) ||
	((c >= '{') && (c <= '~'))) {
	if (mode == PROG)
	    arith_int(c);
	else
	    arith_real(c);
	return 0;
    }
    
    switch (c) {
    
    case 'Y':
    case CHS:
    case PLUS:
    case MINUS:
    case DIVIDE:
    case MULT:
    case AND:
    case OR:
    case XOR:
    case NOT:
    case MODULUS:
    case PERCENT:
    case PERCENTCH:
    case YTOX:
    case SHIFTL:
    case SHIFTR:
    case SHIFTYL:
    case SHIFTYR:
    case PRIMEF:
    case RECI:
	if (mode == PROG)
	    arith_int(c);
	else
	    arith_real(c);
	break;

    case PVAL:
    case FVAL:
    case PMT:
    case INTST:
    case FCLR:
    case NFIN:
    case TIMES12:
    case DIVIDE12:
    case BEGIN:
    case DYS:
    case TDY:
	financial(c);
	break;
    
    case ASCIIM:
    case BIN:
    case IP:
    case OCT:
    case DEC:
    case HEX:

    case FLOAT:
    case ENG:
    case PLACES:

    case FIN:
    case STAT:
    case SCI:
    case PROG:
	change_base(c);
	break;
	
    case ROLLDOWN:
	if (mode == PROG) {
	    tempx = pop(); 
	    t = tempx;
	} else {
	    tempxf = popf();
	    tf = tempxf;
	}
	dispnums();
	break;
	
    case CLX: /* Clear X */
	if (mode == PROG) {
	    l = pop(); 
	    tempx = 0;
	    push(tempx); 
	} else {
	    lf = popf();
	    tempxf = 0.0;
	    pushf(tempxf);
	}
	lift_needed = FALSE;
	dispnums();
	break;

    case CLR: /* Clear registers */
	c = dialog("Clear all registers? (Y or N)");
	if (toupper(c) == 'Y') {
	    for (i = 0; i < NUMREGS; i++) {
		if (mode == PROG) {
		    reg[i] = 0;
		} else {
		    regf[i] = 0.0;
		}
	    }
	}
	break;
	    
    case 'M':
	degree = !degree;
	print_deg();
	break;
	
    case DEGREE:
	degree = 1;
	print_deg();
	msg("Angles are in degrees");
	break;
	
    case RADIAN:
	degree = 0;
	print_deg();
	msg("Angles are in radians");
	break;
	
    case E:
	tempxf = exp(1.0);
	pushf(tempxf);
	dispnums();
	break;

    case PI:
	tempxf = pi;
	tempx = tempxf;
	pushf(tempxf);
	dispnums();
	break;
	
    case SIN:
    case COS:
    case TAN:
    case SINH:
    case COSH:
    case TANH:
    case LOGE:
    case LOG10:
    case SQR:
    case ROOT:
    case CUBE:
    case CROOT:
    case FRC:
    case INT:
    case ETOX:
    case HMS:
    case RTOP:
    case DTOR:
	trig(c);
	break;
	
    case SUMR:
    case SUM0:
    case NSTAT:
    case MEAN:
    case MEANY:
    case S_DEV:
    case SUM:
    case FACT:
    case COMB:
    case PERM:
    case LR:
    case CALCY:
	summation(c);
	break;
	
    case ENTER:
	if (mode == PROG) {
	    tempx = pop(); 
	    push(tempx); 
	    push(tempx); 
	} else {
	    tempxf = popf();
	    pushf(tempxf);
	    pushf(tempxf);
	}
	lift_needed = FALSE;
	dispnums();
	break;
	
    case LASTX:
	if (mode == PROG)
	    push(l); 
	else
	    pushf(lf);
	dispnums();
	break;
	
    case QUIT:
	if (is_resident)
	    c = QUIT;
	else {

#ifndef HAVE_CONFIG_H /* gtk */
	    c = dialog("Really quit? (Y or N)");
	    if (toupper(c) == 'Y')
		terminate();
	    else
		c = 0;
	    clear_msg();
#else
	    terminate();
#endif
	}
	break;
	
    case 'R':
    case RECALL:
	i = recall("Recall from register [0-9]");
	if (isdigit(i)) {
	    char buf[80];
	    i -= '0';
	    if ((i >= 0) && (i < NUMREGS)) {
		if (mode == PROG) {
		    tempx = reg[i];
		    push(tempx); 
		} else {
		    tempxf = regf[i];
		    pushf(tempxf);
		}
		dispnums();
		sprintf(buf, "Recalled from R%d", i);
		msg(buf);
	    }
	}
	break;
	
    case 'S':
    case STORE:
	stop_entering();
	i = store("Store in register [0-9] (or [+-*/] reg.)");
	j = i;
	if ((j=='+') || (j=='-') || (j=='*') || (j=='/'))
	    i = store("Store in register [0-9]");
	else {
	    j = 0;
	    if (isdigit(i))
		i -= '0';
	    else
		break;
	}
	if ((i >= 0) && (i < NUMREGS)) {
	    char op[3], buf[80];

	    switch (j) {
	    case 0:
		reg[i] = x;
		regf[i] = xf;
		strcpy(op, "");
		break;
	    case '+':
		reg[i] += x;
		regf[i] += xf;
		strcpy(op, "+ ");
		break;
	    case '-':
		reg[i] -= x;
		regf[i] -= xf;
		strcpy(op, "- ");
		break;
	    case '*':
		reg[i] *= x;
		regf[i] *= xf;
		strcpy(op, "* ");
		break;
	    case '/':
		reg[i] /= x;
		regf[i] /= xf;
		strcpy(op, "/ ");
		break;
	    }
	    dispreg(i);
	    sprintf(buf, "Stored %sin R%d", op, i);
	    msg(buf);
	}
	break;
	
    case STOPLUS:
    case STOMINUS:
	stop_entering();
	i = store("Store in register [0-9]");
	if (isdigit(i))
	    i -= '0';
	else
	    break;
	if ((i >= 0) && (i < NUMREGS)) {
	    char buf[80];

	    if (c == STOPLUS) {
		reg[i] += x;
		regf[i] += xf;
	    } else {
		reg[i] -= x;
		regf[i] -= xf;
	    }
	    dispreg(i);
	    sprintf(buf, "%s R%d", (c==STOPLUS)? "Added to": "Subtracted from", i);
	    msg(buf);
	}
	break;
	
    case 'X':
    case XNY:
	if (mode == PROG) {
	    tempx = pop(); 
	    tempy = pop(); 
	    push(tempx);   
	    push(tempy);   
	} else {
	    tempxf = popf();
	    tempyf = popf();
	    pushf(tempxf);  
	    pushf(tempyf);  
	}
	dispnums();
	break;

    case XNL:
	if (mode == PROG) {
	    tempx = x;
	    x = l;
	    l = tempx;
	} else {
	    tempxf = xf;
	    xf = lf;
	    lf = tempxf;
	}
	dispnums();
	break;

    case XNT:
	if (mode == PROG) {
	    tempx = x;
	    x = t;
	    t = tempx;
	} else {
	    tempxf = xf;
	    xf = tf;
	    tf = tempxf;
	}
	dispnums();
	break;

    case XNZ:
	if (mode == PROG) {
	    tempx = x;
	    x = z;
	    z = tempx;
	} else {
	    tempxf = xf;
	    xf = zf;
	    zf = tempxf;
	}
	dispnums();
	break;

    case '?':
    case 'H':
    case HELP:
	pop_up_help();
	break;

    case NOP:
	break;

    default:
	put_a_char(BELL);
	break;
    }
    return 0;
}

static void getReg(char *b) {
    long l;
    int i;

    if (sscanf(b, "%d %ld", &i, &l))
	if (i >= 0 && i < NUMREGS)
	    reg[i] = l;
}

static void getRegf(char *b) {
    double l;
    int i;

    if (sscanf(b, "%d %lg", &i, &l))
	if (i >= 0 && i < NUMREGS)
	    regf[i] = l;
}

static void getMode(char *b) {
    mode = 
	(strcmp(b, "sci")==0)  ? SCI:
	(strcmp(b, "fin")==0)  ? FIN:
	(strcmp(b, "stat")==0) ? STAT:
	(strcmp(b, "prog")==0) ? PROG: SCI;
}

static void getIntmode(char *b) {
    intmode = 
	strcmp(b, "asci")==0? ASCIIM:
	strcmp(b, "ip")==0? IP:
	strcmp(b, "bin")==0? BIN:
	strcmp(b, "oct")==0? OCT:
	strcmp(b, "dec")==0? DEC:
	strcmp(b, "hex")==0? HEX: DEC;
}

static void getFloatmode(char *b) {
    floatmode = 
	strcmp(b, "FLT")==0? FLOAT:
	strcmp(b, "ENG")==0? ENG: FLOAT;
}
	
static void readSettings() {
    FILE *f = NULL;
    char *homeDir = getenv("HOME");
    char fileName[200], buf[80];
    if (homeDir) {
	strcpy(fileName, homeDir);
	if (fileName[strlen(fileName) - 1] != '/')
	    strcat(fileName, "/");
	strcat(fileName, ".dcalcrc");
	if ((f = fopen(fileName, "r")) != NULL) {
	    while (fgets(buf, 80, f) != NULL) {
		char *b = buf;

		if (*b && buf[strlen(buf) - 1] == '\n')
		    buf[strlen(buf) - 1] = 0;

		if (*b == '#')
		    continue;

		while (!isspace(*b))
		    b++;
		*b++ = 0;
		if (strcmp(buf, "x") == 0) x = atol(b);
		if (strcmp(buf, "y") == 0) y = atol(b);
		if (strcmp(buf, "z") == 0) z = atol(b);
		if (strcmp(buf, "t") == 0) t = atol(b);
		if (strcmp(buf, "l") == 0) l = atol(b);
		if (strcmp(buf, "xf") == 0) xf = atof(b);
		if (strcmp(buf, "yf") == 0) yf = atof(b);
		if (strcmp(buf, "zf") == 0) zf = atof(b);
		if (strcmp(buf, "tf") == 0) tf = atof(b);
		if (strcmp(buf, "lf") == 0) lf = atof(b);
		if (strcmp(buf, "reg") == 0) getReg(b);
		if (strcmp(buf, "regf") == 0) getRegf(b);
		if (strcmp(buf, "angularMode") == 0) degree = atoi(b);
		if (strcmp(buf, "places") == 0) decplaces = atoi(b);
		if (strcmp(buf, "mode") == 0) getMode(b);
		if (strcmp(buf, "intmode") == 0) getIntmode(b);
		if (strcmp(buf, "floatmode") == 0) getFloatmode(b);
		if (strcmp(buf, "gui") == 0) readGuiSettings(b);
		if (strcmp(buf, "annuitymode") == 0) finPayAt0 = atol(b);
	    }
	    fclose(f);
	}
    }
}

static FILE *saveSettings() {
    FILE *f = NULL;
    char *homeDir = getenv("HOME");
    char fileName[200], buf[80];
    char lFormat[] = "%s %.25lg\n";
    char dFormat[] = "%s %ld\n";
    char iFormat[] = "%s %d\n";
    int i;

    stop_entering();
    if (homeDir) {
	strcpy(fileName, homeDir);
	if (fileName[strlen(fileName) - 1] != '/')
	    strcat(fileName, "/");
	strcat(fileName, ".dcalcrc");
	if ((f = fopen(fileName, "w")) != NULL) {
	    fprintf(f, "# %s\n", DEF_SIG(VERSION));
	    fprintf(f, "# don't edit this file - dcalc will overwrite it\n");
	    fprintf(f, dFormat, "x", x);
	    fprintf(f, dFormat, "y", y);
	    fprintf(f, dFormat, "z", z);
	    fprintf(f, dFormat, "t", t);
	    fprintf(f, dFormat, "l", l);
	    fprintf(f, lFormat, "xf", xf);
	    fprintf(f, lFormat, "yf", yf);
	    fprintf(f, lFormat, "zf", zf);
	    fprintf(f, lFormat, "tf", tf);
	    fprintf(f, lFormat, "lf", lf);
	    for (i = 0; i < NUMREGS; i++) {
		sprintf(buf, "reg %d", i);
		fprintf(f, dFormat, buf, reg[i]);
	    }
	    for (i = 0; i < NUMREGS; i++) {
		sprintf(buf, "regf %d", i);
		fprintf(f, lFormat, buf, regf[i]);
	    }
	    fprintf(f, iFormat, "angularMode", degree);
	    fprintf(f, iFormat, "places", decplaces);
	    fprintf(f, "mode %s\n", 
		    mode==SCI? "sci":
		    mode==FIN? "fin":
		    mode==STAT? "stat":
		    mode==PROG? "prog":"sci");
	    fprintf(f, "intmode %s\n", 
		    intmode==ASCIIM? "asci":
		    intmode==IP? "ip":
		    intmode==BIN? "bin":
		    intmode==OCT? "oct":
		    intmode==DEC? "dec":
		    intmode==HEX? "hex": "dec");
	    fprintf(f, "floatmode %s\n",
		    floatmode==FLOAT? "FLT": "ENG");
	    saveGuiSettings(f);
	    /* Rev 2.7 on: */
	    fprintf(f, "annuitymode %d\n", finPayAt0? 1: 0);
	    fclose(f);
	}
    }
    return f;
}

void initialise(void) {
    int	i;
    
    if (!was_initialised) {
	was_initialised = 1;
	os_init();
	
	x = y = z = t = l = 0;
	xf = yf = zf = tf = lf = 0.0;
	decplaces = 2;
	mode = SCI;
	intmode = DEC;
	floatmode = FLOAT;
	entering = FALSE;
	lift_needed = TRUE;
	invert = FALSE;
	degree = TRUE;
	last_was_fin_key = FALSE;
	inptr = inbuf;
	*inptr = ASCIIZ;
	max_digits = 20;
	pi = 4.0 * atan(1.0);
	d_to_r = pi / 180.0;
	r_to_d = 180.0 / pi;
	
	for (i = 0; i < NUMREGS; i++) {
	    reg[i] = 0;
	    regf[i] = 0.0;
	}

	readSettings();
	if ((mode == PROG) && (intmode == ASCIIM))
	    os_raw_mode(1);
    }
}

extern void terminate() {
    saveSettings();
    os_term();
}


/* For emacs: */

/* Local Variables: */
/* eval:(setq tab-width 8) */
/* End: */
