/*
  main.c: gtk driver for dcalc

    Contact info:
    bhepple@bit.net.au
    http://www.bit.net.au/~bhepple/dcalc

    Copyright (C) 1999,2000 Bob Hepple

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.

 */

/* With glade 0.5.11 that every time the menubar is edited, one item
   in every radio group is turned on and interface.c gets a
   gtk_check_menu_item_set_active for that item. This results in
   seemingless harmless but distracting runtime errors. */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "interface.h"
#include "support.h"
#include "dcalc.h"
#include "main.h"

/* naughty global exported to callbacks.c: */
GtkWidget *main_window;
GtkWidget *mem_dialog;
GtkWidget *convert;
GtkWidget *registers = NULL;
int mem_value;
int raw_mode = 0;
int is_resident = 0;
dcalc_button *l_buttons, *r_buttons;
char calculatorFont[100];

dcalc_button sci_buttons[] = {
  { 0, 0, SIN, "SIN", "Sine of X (or INV)" },
  { 0, 1, COS, "COS", "Cosine of X (or INV)" },
  { 0, 2, TAN, "TAN", "Tangent of X (or INV)" },
  { 0, 3, HMS, "h.ms", "Convert X to Hours.MinsSecs (or INV)" },
  { 0, 4, RTOP, "r->p", "Convert (X,Y) to polar coordinates (or INV)" },
  { 0, 5, DTOR, "d->r", "Convert degrees to radians (or INV)" },
  { 0, 6, LOG10, "LOG", "Logarithm (base 10) (or INV)" },
  { 0, 7, LOGE, "LN", "Natural logarithm (base e) (or INV)" },
  { 1, 0, SINH, "SINh", "Hyperbolic sine (or INV)" },
  { 1, 1, COSH, "COSh", "Hyperbolic cosine (or INV)" },
  { 1, 2, TANH, "TANh", "Hyperbolic tangent (or INV)" },
  { 1, 3, NOP, " ", " " },
  { 1, 4, NOP, " ", " " },
  { 1, 5, PI, "PI", "PI" },
  { 1, 6, INT, "INT", "Integer part of X" },
  { 1, 7, FRC, "FRC", "Fractional part of X" },
  { 2, 0, INV, "INV", "Invert functions" },
  { 2, 1, NOP, " ", " " },
  { 2, 2, NOP, " ", " " },
  { 2, 3, NOP, " ", " " },
  { 2, 4, NOP, " ", " " },
  { 2, 5, E, "e", "e" },
  { 2, 6, ROOT, "SQRT", "Square root of X" },
  { 2, 7, SQR, "x2", "X squared" },
  { -1, -1, 0, " ", " " }
};

dcalc_button fin_buttons[] = {
  { 0, 0, NFIN, "n", "Number of periods (or INV)" },
  { 0, 1, INTST, "i", "Interest rate in % per period (or INV)" },
  { 0, 2, PVAL, "PV", "Present value (or INV)" },
  { 0, 3, PMT, "PMT", "Payment (or INV)" },
  { 0, 4, FVAL, "FV", "Future value (or INV)" },
  { 0, 5, FCLR, "CLf", "Clear compound interest registers (5-9)" },
  { 0, 6, LOG10, "LOG", "Logarithm (base 10) (or INV)" },
  { 0, 7, LOGE, "LN", "Natural logarithm (base e)" },
  { 1, 0, TIMES12, "*12", "X times 12" },
  { 1, 1, DIVIDE12, "/12", "X divided by 12" },
  { 1, 2, BEGIN, "B/E", "Toggle Annuity at beginning of period" },
  { 1, 3, DYS, "DYS", "Days between dates" },
  { 1, 4, TDY, "TDY", "Today's date" },
  { 1, 5, PI, "PI", "PI" },
  { 1, 6, INT, "INT", "Integer part of X" },
  { 1, 7, FRC, "FRC", "Fractional part of X" },
  { 2, 0, INV, "INV", "Invert functions" },
  { 2, 1, NOP, " ", " " },
  { 2, 2, NOP, " ", " " },
  { 2, 3, NOP, " ", " " },
  { 2, 4, NOP, " ", " " },
  { 2, 5, E, "e", "e" },
  { 2, 6, ROOT, "SQRT", "Square root of X" },
  { 2, 7, SQR, "x2", "X squared" },
  { -1, -1, 0, " ", " " }
};
  
dcalc_button stat_buttons[] = {
  { 0, 0, SUM, "SUM+", "Add X to the accumulator (register 7)" },
  { 0, 1, MEAN, "m(X)", "Calculate the mean of X" },
  { 0, 2, MEANY, "m(Y)", "Calculate the mean of Y" },
  { 0, 3, S_DEV, "StdD", "Calculate the standard deviations" },
  { 0, 4, NSTAT, "N", "Number of samples" },
  { 0, 5, FACT, "x!", "X factorial" },
  { 0, 6, LOG10, "LOG", "Logarithm (base 10)" },
  { 0, 7, LOGE, "LN", "Natural logarithm (base e)" },
  { 1, 0, SUMR, "SUM-", "subtract X from the accumulator (register 7)" },
  { 1, 1, PERM, "yPx", "x permutations of y" },
  { 1, 2, COMB, "yCx", "x combinations of y" },
  { 1, 3, LR, "LR", "Linear Regression" },
  { 1, 4, CALCY, "f(x)", "Calculate Y from X using linear regression results (Inv: calculate X from Y)" },
  { 1, 5, PI, "PI", "PI" },
  { 1, 6, INT, "INT", "Integer part of X" },
  { 1, 7, FRC, "FRC", "Fractional part of X" },
  { 2, 0, INV, "INV", "Invert functions" },
  { 2, 1, NOP, " ", " " },
  { 2, 2, NOP, " ", " " },
  { 2, 3, NOP, " ", " " },
  { 2, 4, NOP, " ", " " },
  { 2, 5, E, "e", "e" },
  { 2, 6, ROOT, "SQRT", "Square root of X" },
  { 2, 7, SQR, "x2", "X squared" },
  { -1, -1, 0, " ", " " }
};
  
dcalc_button prog_buttons[] = {
  { 0, 0, BIN, "BIN", "Binary mode" },
  { 0, 1, OCT, "OCT", "Octal mode" },
  { 0, 2, DEC, "DEC", "Decimal mode" },
  { 0, 3, HEX, "HEX", "Hexadecimal mode" },
  { 0, 4, ASCIIM, "ASC", "Ascii mode" },
  { 0, 5, IP, "IP", "IP address mode" },
  { 0, 6, NOP, " ", " " },
  { 0, 7, NOP, " ", " " },
  { 1, 0, AND, "AND", "y and y" },
  { 1, 1, OR, "OR", "y or x" },
  { 1, 2, XOR, "XOR", "y xor x" },
  { 1, 3, MODULUS, "MOD", "y mod x" },
  { 1, 4, NOT, "~X", "not X (complement)" },
  { 1, 5, SHIFTYL, "y<<x", "Shift Y, X bits left" },
  { 1, 6, SHIFTYR, "y>>x", "Shift Y, X bits right" },
  { 1, 7, NOP, " ", " " },
  { 2, 0, PRIMEF, "PF", "Prime factors of X" },
  { 2, 1, NOP, " ", " " },
  { 2, 2, NOP, " ", " " },
  { 2, 3, NOP, " ", " " },
  { 2, 4, NOP, " ", " " },
  { 2, 5, SHIFTL, "<<", "Shift X left 1 bit" },
  { 2, 6, SHIFTR, ">>", "Shift Y right 1 bit" },
  { 2, 7, NOP, " ", " " },
  { -1, -1, 0, " ", " " }
};

dcalc_button float_buttons[] = {
  { 0, 0, FIN, "FIN", "Finance mode" },
  { 0, 1, STAT, "STA", "Statistics mode" },
  { 0, 2, PROG, "PRO", "Programmers mode" },
  { 0, 3, NOP, " ", " " },
  { 0, 4, 'E', "E", " " },
  { 0, 5, PERCENT, "%", "X percent of Y (or INV)" },
  { 1, 0, NOP, " ", " " },
  { 1, 1, NOP, " ", " " },
  { 1, 2, NOP, " ", " " },
  { 1, 3, YTOX, "y^X", "Y to the power of X" },
  { 1, 4, RECI, "1/X", "Reciprocal of X" },
  { 1, 5, PERCENTCH, "%CH", "Percentage change from Y to X (or INV)" },
  { 2, 0, FLOAT, "FLT", "Floating point notation" },
  { 2, 1, ENG, "ENG", "Engineering notation" },
  { 2, 2, NOP, " ", " " },
  { 2, 3, '7', "7", " " },
  { 2, 4, '8', "8", " " },
  { 2, 5, '9', "9", " " },
  { 3, 0, STORE, "STO", "Store X to a register" },
  { 3, 1, STOPLUS, "ST+", "Add X to a register" },
  { 3, 2, STOMINUS, "ST-", "Subtract X from a register" },
  { 3, 3, '4', "4", " " },
  { 3, 4, '5', "5", " " },
  { 3, 5, '6', "6", " " },
  { 4, 0, RECALL, "RCL", "Recall from a register" },
  { 4, 1, CLR, "CLr", "Clear all registers" },
  { 4, 2, NOP, " ", " " }, /* reserved for CLu */
  { 4, 3, '1', "1", " " },
  { 4, 4, '2', "2", " " },
  { 4, 5, '3', "3", " " },
  { 5, 0, ROLLDOWN, "ROLL", "Roll stack down" },
  { 5, 1, CLX, "CLx", "Clear X" },
  { 5, 2, BACKSPACE, "BSP", "Backspace" },
  { 5, 3, '0', "0", " " },
  { 5, 4, '.', ".", " " },
  { 5, 5, CHS, "CHS", "Change sign of X" },
  { -1, -1, 0, " ", " " }
};

dcalc_button int_buttons[] = {
  { 0, 0, SCI, "SCI", "Scientific mode" },
  { 0, 1, FIN, "FIN", "Finance mode" },
  { 0, 2, STAT, "STA", "Statistics mode" },
  { 0, 3, 'D', "D", " " },
  { 0, 4, 'E', "E", " " },
  { 0, 5, 'F', "F", " " },
  { 1, 0, NOP, " ", " " },
  { 1, 1, NOP, " ", " " },
  { 1, 2, NOP, " ", " " },
  { 1, 3, 'A', "A", " " },
  { 1, 4, 'B', "B", " " },
  { 1, 5, 'C', "C", " " },
  { 2, 0, NOP, " ", " " },
  { 2, 1, NOP, " ", " " },
  { 2, 2, NOP, " ", " " },
  { 2, 3, '7', "7", " " },
  { 2, 4, '8', "8", " " },
  { 2, 5, '9', "9", " " },
  { 3, 0, STORE, "STO", "Store X to a register" },
  { 3, 1, STOPLUS, "ST+", "Add X to a register" },
  { 3, 2, STOMINUS, "ST-", "Subtract X from a register" },
  { 3, 3, '4', "4", " " },
  { 3, 4, '5', "5", " " },
  { 3, 5, '6', "6", " " },
  { 4, 0, RECALL, "RCL", "Recall from a register" },
  { 4, 1, CLR, "CLr", "Clear all registers" },
  { 4, 2, NOP, " ", " " },
  { 4, 3, '1', "1", " " },
  { 4, 4, '2', "2", " " },
  { 4, 5, '3', "3", " " },
  { 5, 0, ROLLDOWN, "ROLL", "Roll stack down" },
  { 5, 1, CLX, "CLx", "Clear X" },
  { 5, 2, BACKSPACE, "BSP", "Backspace" },
  { 5, 3, '0', "0", " " },
  { 5, 4, '.', ".", " " },
  { 5, 5, CHS, "CHS", "Change sign of X" },
  { -1, -1, 0, " ", " " }
};

int button_command(dcalc_button *b, int i, int j) {
  while (b != NULL && b->i >= 0)
    if (b->i == i && b->j == j)
      return(b->command);
    else
      b++;
  return(NOP);
}

static void display_panel(char which_panel, dcalc_button *b) {
	gpointer g;
	char widget_name[80];

	if (which_panel == 'R')
		r_buttons = b;
	else
		l_buttons = b;

	while (b != NULL && b->i >= 0) {
		sprintf(widget_name, "%c%d_%d", which_panel, b->i, b->j);
		g = gtk_object_get_data(GTK_OBJECT (main_window), widget_name);
		if (g) {
			GtkWidget *child;
			GtkTooltipsData *tooltipsdata;

			child = GTK_BIN (g)->child;
			if (GTK_IS_LABEL (child))
				gtk_label_set_text (GTK_LABEL (child), b->label);
			tooltipsdata = gtk_tooltips_data_get(g);
			if (tooltipsdata)
				gtk_tooltips_set_tip(tooltipsdata->tooltips, g, b->tooltip, NULL);
			else
				fprintf(stderr, "Can't find tooltip for widget %s\n", widget_name);
		} else
			fprintf(stderr, "Can't find widget %s\n", widget_name);
		b++;
	}
}

void display_buttons(int c) {
	GtkWidget *w;

	if (c == PROG) {
		display_panel('R', int_buttons);
		display_panel('L', prog_buttons);
	} else {
		switch (c) {
		case STAT:
			float_buttons[0].command = PROG;
			strcpy(float_buttons[0].label, "PRO");
			strcpy(float_buttons[0].tooltip, "Programmers mode");
			float_buttons[1].command = SCI;
			strcpy(float_buttons[1].label, "SCI");
			strcpy(float_buttons[1].tooltip, "Scientific mode");
			float_buttons[2].command = FIN;
			strcpy(float_buttons[2].label, "FIN");
			strcpy(float_buttons[2].tooltip, "Finance mode");
			display_panel('L', stat_buttons);	
			break;
		case FIN:
			float_buttons[0].command = STAT;
			strcpy(float_buttons[0].label, "STA");
			strcpy(float_buttons[0].tooltip, "Statistics mode");
			float_buttons[1].command = PROG;
			strcpy(float_buttons[1].label, "PRO");
			strcpy(float_buttons[1].tooltip, "Programmers mode");
			float_buttons[2].command = SCI;
			strcpy(float_buttons[2].label, "SCI");
			strcpy(float_buttons[2].tooltip, "Scientific mode");
			display_panel('L', fin_buttons);
			break;
		default: /* SCI */
			float_buttons[0].command = FIN;
			strcpy(float_buttons[0].label, "FIN");
			strcpy(float_buttons[0].tooltip, "Finance mode");
			float_buttons[1].command = STAT;
			strcpy(float_buttons[1].label, "STA");
			strcpy(float_buttons[1].tooltip, "Statistics mode");
			float_buttons[2].command = PROG;
			strcpy(float_buttons[2].label, "PRO");
			strcpy(float_buttons[2].tooltip, "Programmers mode");
			display_panel('L', sci_buttons);
			break;
		}
		display_panel('R', float_buttons);
	}      
	w = gtk_object_get_data(GTK_OBJECT(main_window), "degree");
	if (w)
		gtk_widget_set_sensitive(w, (c==PROG)? 0: 1);
	else
		fprintf(stderr, "gdcalc: can't find degree menu item\n");

	w = gtk_object_get_data(GTK_OBJECT(main_window), "radians");
	if (w)
		gtk_widget_set_sensitive(w, (c==PROG)? 0: 1);
	else
		fprintf(stderr, "gdcalc: can't find radians menu item\n");

	w = gtk_object_get_data(GTK_OBJECT(main_window), "annuity_in_advance");
	if (w)
		gtk_widget_set_sensitive(w, (c==FIN)? 1: 0);
	else
		fprintf(stderr, "gdcalc: can't find annuity menu item\n");

	w = gtk_object_get_data(GTK_OBJECT(main_window), "convert");
	if (w)
		gtk_widget_set_sensitive(w, (c==PROG)? 0: 1);
	else
		fprintf(stderr, "gdcalc: can't find convert menu item\n");

	w = gtk_object_get_data(GTK_OBJECT(main_window), "floating_point");
	if (w)
		gtk_widget_set_sensitive(w, (c==PROG)? 0: 1);
	else
		fprintf(stderr, "gdcalc: can't find floating point menu item\n");

	w = gtk_object_get_data(GTK_OBJECT(main_window), "engineering");
	if (w)
		gtk_widget_set_sensitive(w, (c==PROG)? 0: 1);
	else
		fprintf(stderr, "gdcalc: can't find engineering menu item\n");

	w = gtk_object_get_data(GTK_OBJECT(main_window), "places");
	if (w)
		gtk_widget_set_sensitive(w, (c==PROG)? 0: 1);
	else
		fprintf(stderr, "gdcalc: can't find places menu item\n");
}

static void print_number(char *buf, char *name) 
{
	gpointer g;
	g = gtk_object_get_data(GTK_OBJECT (main_window), name);

	gtk_entry_set_text (GTK_ENTRY (g), buf);
}

void print_x(char *buf) {
	print_number(buf, "entry_X");
}

void dispnums()
{
    char 	outbuf[80];

    if (entering) {
		strcpy(outbuf, inbuf);
    } else
		fmt_base(outbuf, x, xf);

    prep_for_output(outbuf);
    print_x(outbuf);
    
    fmt_base(outbuf, y, yf);
    prep_for_output(outbuf);
    print_number(outbuf, "entry_Y");

    fmt_base(outbuf, z, zf);
    prep_for_output(outbuf);
    print_number(outbuf, "entry_Z");

    fmt_base(outbuf, t, tf);
    prep_for_output(outbuf);
    print_number(outbuf, "entry_T");

    fmt_base(outbuf, l, lf);
    prep_for_output(outbuf);
    print_number(outbuf, "entry_L");
}

/* only used for the bell ... */
void put_a_char(int i) {
	gdk_beep();
}

static char *printBase() {
	if (mode == PROG)
		switch (intmode) {
		case ASCIIM:
			return("prog:asc: ");
		case BIN:
			return("prog:bin: ");
		case OCT:
			return("prog:oct: ");
		case DEC:
			return("prog:dec: ");
		case HEX:
			return("prog:hex: ");
		} 
	else
		if (l_buttons == sci_buttons)
			return "sci: ";
		else if (l_buttons == fin_buttons)
			return "fin: ";
		else
			return "stats: ";
	return "";
}

static char *printInv() {
    if (invert)
		return("inv: ");
    else
		return("");
}

static char *printDeg(){
    if (degree)
		return("deg: ");
    else
		return("rad: ");
}

void msg(char *buf) {
	gint context_id;
	gpointer status_bar;
	char status[200];

	strcpy(status, printBase());
	if (mode != PROG) {
		strcat(status, printDeg());
		strcat(status, printInv());
	}
	strcat(status, " ");
	strcat(status, buf);
	status_bar = gtk_object_get_data(GTK_OBJECT (main_window), _("statusbar"));
	context_id = gtk_statusbar_get_context_id(GTK_STATUSBAR(status_bar), 
											  "Statusbar example");
	gtk_statusbar_push( GTK_STATUSBAR(status_bar), context_id, status);
}

void clear_msg() {
	msg("");
}

/* These not needed for GTK version: */
void prinbase() {};
void print_string(char *buf) {}
void print_inv() {}
void print_deg(){}
void pop_up_help(void) {}
void os_raw_mode(int i) { raw_mode = i; }
void os_init(void) {}
void os_term(void) {gtk_main_quit();}
int get_a_char(int *c) {return 0;}

int dialog(char *buf) {
	GtkWidget *dialog_popup;
	gpointer g;
	int reply;

	dialog_popup = create_gen_dialog ();
	g = gtk_object_get_data(GTK_OBJECT (dialog_popup), _("msg"));
	gtk_label_set_text(GTK_LABEL(g), buf);
	reply = gnome_dialog_run(GNOME_DIALOG(dialog_popup));
	gtk_widget_destroy(dialog_popup);
	return (reply==0)? 'Y': 'N';
}

int places(char *buf) {
	GtkWidget *places_dialog;
	GtkSpinButton *spinbutton;
	int reply;
	gfloat decplacesf = decplaces;

	places_dialog = create_places ();
	spinbutton = gtk_object_get_data(GTK_OBJECT (places_dialog), _("places_spinbutton"));
	gtk_spin_button_set_value(spinbutton, decplacesf);

	reply = gnome_dialog_run(GNOME_DIALOG(places_dialog));
	if (reply == 0)
		reply = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(spinbutton));
	else
		reply = -1;

	gtk_widget_destroy(places_dialog);

	return reply;
}

void fixFont(GtkWidget *main_window, char *name) {
	GdkFont *fixed_font;
	GtkStyle *style;
	GtkEntry *entry;

	fixed_font = gdk_font_load(calculatorFont);
	if (fixed_font) {
		entry = gtk_object_get_data(GTK_OBJECT (main_window), name);
		style = gtk_widget_get_style(GTK_WIDGET(entry));
		gdk_font_unref (style->font);
		style->font = fixed_font;
		gdk_font_ref (style->font);
		gtk_widget_set_style (GTK_WIDGET(entry), style);
	}
}

static void xdispreg(GtkWidget *dialog, int i) {
	char labelname[80], outbuf[80];
	GtkWidget *entry;

	if (dialog) {
		sprintf(labelname, "mem%d", i);
		fixFont(dialog, labelname);
		entry = gtk_object_get_data(GTK_OBJECT (dialog), labelname);
		fmt_base(outbuf, reg[i], regf[i]);
		prep_for_output(outbuf);
		if (l_buttons == fin_buttons)
			switch (i) {
			case FIN_NUM_REG: strcat(outbuf, " = periods"); break;
			case FIN_INT_REG: strcat(outbuf, " = interest (%)"); break;
			case FIN_PVAL_REG: strcat(outbuf, " = PVal"); break;
			case FIN_PMT_REG: strcat(outbuf, " = payment"); break;
			case FIN_FVAL_REG: strcat(outbuf, " = FVal"); break;
			}
		else if (l_buttons == stat_buttons)
			switch (i) {
			case STAT_R2: strcat(outbuf, " = correlation (1=good 0=poor)"); break;
			case STAT_Y_INT: strcat(outbuf, " = Y intersect"); break;
			case STAT_SLOPE: strcat(outbuf, " = slope"); break;
			case STAT_NUM_REG: strcat(outbuf, " = n"); break;
			case STAT_SUMX_REG: strcat(outbuf, " = sum of X's"); break;
			case STAT_SUMX2_REG: strcat(outbuf, " = sum of x2's"); break;
			case STAT_SUMY_REG: strcat(outbuf, " = sum of Y's"); break;
			case STAT_SUMY2_REG: strcat(outbuf, " = sum of y2's"); break;
			case STAT_SUMXY_REG: strcat(outbuf, " = sum of XY's"); break;
			}

		gtk_entry_set_text(GTK_ENTRY(entry), outbuf);
	}
}

void dispreg(int i) {
	xdispreg(registers, i);
}

void dispregs(void) {
	int i;

	if (registers)
		for (i = 0; i < NUMREGS; i++)
			dispreg(i);
}

void toggle_registers() {
	if (!registers) {
		registers = create_registers();
		dispregs();
		gtk_widget_show(registers);
	} else {
		gtk_widget_destroy(registers);
		registers = NULL;
	}
}

void pop_up_reg(void) {
	toggle_registers();
}

static int do_mem_dialog(char *buf) {
	int reply, i;

	mem_dialog = create_mem_dialog ();
	gtk_window_set_title (GTK_WINDOW (mem_dialog), buf);
	for (i=0; i < NUMREGS; i++)
		xdispreg(mem_dialog, i);
	reply = gnome_dialog_run(GNOME_DIALOG(mem_dialog));
	gnome_dialog_close(GNOME_DIALOG(mem_dialog));
	return (reply==0)? -1: mem_value+'0';
}

int store(char *buf) {
	return(do_mem_dialog("Store..."));
}

int recall(char *buf) {
	return(do_mem_dialog("Recall..."));
}

int
main (int argc, char *argv[])
{
	char *cfont = NULL;
	struct poptOption options[] = {
		{"calculatorFont", 'f', POPT_ARG_STRING, NULL, 0, "Font for numbers", "FONT"},
		{NULL, '\0', 0, NULL, 0, NULL, NULL} /* end the list */
	};
	GtkWidget *w;

	bindtextdomain (PACKAGE, PACKAGE_LOCALE_DIR);
	textdomain (PACKAGE);

	options[0].arg = &cfont;
	gnome_init_with_popt_table("gdcalc", VERSION, argc, argv, options, 0, NULL);

	if (cfont && *cfont)
		strcpy(calculatorFont, cfont);
	else
		strcpy(calculatorFont, "-misc-fixed-medium-r-*-*-14-*-*-*-*-*-*-*");  

	main_window = create_main_window ();
	/* Load a fixed font for the numeric displays: */
	fixFont(main_window, "entry_X");
	fixFont(main_window, "entry_Y");
	fixFont(main_window, "entry_Z");
	fixFont(main_window, "entry_T");
	fixFont(main_window, "entry_L");
  
	gtk_widget_show (main_window);
	initialise();
	display_buttons(mode);

	/* gtk_check_menu_item_set_active actually calls the widgets activate method - ie 
	   simulating a button release!! This can lead to an infinite loop of invocations */
	w = gtk_object_get_data(GTK_OBJECT(main_window), 
							(mode==SCI)? "scientific":
							(mode==FIN)? "financial":
							(mode==STAT)? "statistics":
							"programming");
	if (w)
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), 1);
	else
		printf("Can't find main mode item in menu\n");

	w = gtk_object_get_data(GTK_OBJECT(main_window), 
							degree? "degree":
							"radians");
	if (w)
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), 1);
	else
		printf("Can't find degree/radians item in menu\n");

	w = gtk_object_get_data(GTK_OBJECT(main_window), (floatmode==ENG)? "engineering":
		"floating_point");
	if (w)
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), 1);
	else
		printf("Can't find floating point item in menu\n");

	w = gtk_object_get_data(GTK_OBJECT(main_window), "annuity_in_advance");
	if (w)
		gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(w), finPayAt0? 1: 0);
	else
		printf("Can't find annuity item in menu\n");

	dispnums();
	msg(DEF_SIG(VERSION));

	if (registers) { /* set to 1 by readGuiSettings */
		registers = NULL;
		toggle_registers();
	}

	gtk_main ();
	return 0;
}


/* For emacs: */

/* Local Variables: */
/* eval:(setq tab-width 4) */
/* End: */
